
#ifndef PTOLEMY
#include <stdio.h>
#include "polis.h"
#endif

#ifdef DEBUG
#define STAT  /* enable statistics */
#endif

/* Global constants */

#define OK                      0
#define ERR                     1
#define ILLEGAL                '?'
#define ILLEGALNUM             -1

#define LRU_POLICY             'l'     /* replacement policies */
#define FIFO_POLICY            'f'
#define RANDOM_POLICY          'r'

#define DEMAND                 'd'     /* fetch policies */
#define ALWAYSPREFETCH         'a'
#define LOADFORWARDPREFETCH    'l'
#define SUBBLOCKPREFETCH       'S'
#define MISSPREFETCH           'm'
#define TAGGEDPREFETCH         't'

#define WRITETHROUGH           'w'     /* write policies */
#define COPYBACK               'c'
#define NOWRITEALLOCATE        'n'
#define WRITEALLOCATE          'w'

#ifndef NULL
#define NULL 0
#endif
#ifndef TRUE
#define TRUE 1
#endif 
#ifndef FALSE
#define FALSE 0
#endif 

#ifndef BUFSIZE
#define BUFSIZE 256
#endif 

/* default values */
#define DEFAULT_CACHESIZE 128 
#define DEFAULT_LINESIZE 2
#define DEFAULT_ASSOCIATIVITY  1         /* 1 => direct mapped */ 
#define DEFAULT_HISTORY_LENGTH 8 
#define DEFAULT_THRESHOLD 0.1
#define DEFAULT_CSIMULSTYLE FAST
#define DEFAULT_I_FETCHSIZE 2
#define DEFAULT_MISS_PENALTY 4

/* cache simulation styles */
#define FAST 0
#define FULL 1
#define FULLFAST 2

/* cache parameters */
#define CACHESIZE get_cache_size()        /* in bytes */
#define LINESIZE  get_line_size()         /* in bytes */
#define ASSOCIATIVITY get_associativity() 
#define HISTORY_LENGTH  get_history_length()  /* depth inter-task analysis */
#define THRESHOLD get_threshold() /* (float) min. conflict value to consider */
#define CSIMULSTYLE get_csimulstyle()     /* cache simulation style */
#define I_FETCH_SIZE get_i_fetchsize()    /* average i_fetch size in bytes */
#define MISS_PENALTY get_miss_penalty()   /* n. cycles to resolve a miss */ 

#define NUM_CACHE_LINES  ( CACHESIZE/LINESIZE )
#define NUM_CACHE_SETS   ( NUM_CACHE_LINES/ASSOCIATIVITY )

#define LINESIZE_MASK ( ~(LINESIZE-1) )

/* layout table */
#define KEY_TASK_BITS 3
#define KEY_PATH_BITS 9
#define KEY_TASK_MASK ((1<<KEY_TASK_BITS)-1)
#define KEY_PATH_MASK ((1<<KEY_PATH_BITS)-1)
#define LAYOUT_TABLE_DIM (1<<(KEY_TASK_BITS+KEY_PATH_BITS))
#define NUM_MAX_TASKS  100 

/* inter conflicts table */
#define KEY_TASK_CONFL_BITS 2
#define KEY_PATH_CONFL_BITS 6
#define KEY_TASK_CONFL_MASK ((1<<KEY_TASK_CONFL_BITS)-1)
#define KEY_PATH_CONFL_MASK ((1<<KEY_PATH_CONFL_BITS)-1)
#define CONFL_TABLE_DIM (1<<((KEY_TASK_CONFL_BITS+KEY_PATH_CONFL_BITS)*2))

/* dynamic path_id computation */
#define SET_BIT_PATH_ID( node ) { path_id ^= (1<<node); }

/* dynamic path accumulation */
#ifdef UNIX
#define ADD_NODE_IN_PATH( node )  { add_node_in_path( header_path, node, is_SW, csimulflag ); }
#endif /* UNIX */


/*****************************************************************************
 * Definitions for the compatibility between the simulation in os_sim (UNIX) *
 * and in  PTOLEMY                                                           *
 *****************************************************************************/

#ifdef UNIX

#define CACHE_PATH_LOOKUP \
        cache_path_lookup( task_id, path_id, sgraph, \
            header_path, curr_layout, is_SW, csimulstyle, csimulflag )

#define CACHE(node) \
        cache( *header_path, *sgraph, node, path_found, \
            curr_layout, is_SW, csimulstyle, csimulflag, tracegenflag )

#define CACHE_LAYOUT \
        cache_layout( task_id, path_id, curr_layout, header_path, \
                    sgraph, path_found, is_SW, csimulstyle, csimulflag )

#define CACHE_PLOT_HIT_RATE(max_label) \
        cache_plot_hit_rate( task_id, sgraph, header_path, \
                    max_label, is_SW, csimulflag, csimulstyle )

#define COMPUTE_ABSTRACT_CACHE_STATE \
        compute_abstract_cache_state( task_id, path_id, sgraph, \
            &header_path, &curr_layout, is_SW, csimulstyle, csimulflag, \
            tracegenflag, fptraces )

#define ALLOC_HEADER_PATH( numnodes )  \
        alloc_header_path( numnodes )


#endif /* UNIX */

extern int is_SW;           /* defined in os.c (for os_sim) */
extern int csimulflag;      /*    "       "      "   "  "   */
extern int tracegenflag;    /*    "       "      "   "  "   */ 
/* INSTRUCTION MACRO */
#define _AddInstrFetch(v)   { tmp_i = (v); \
                              n_instr_fetch_path += tmp_i; \
                              n_instr_fetch += tmp_i; }
#define _GetInstrFetch      n_instr_fetch
#define _AddInstrHit(v)     { tmp_f = (v); \
                              n_instr_hit_path += tmp_f; \
                              n_instr_hit += tmp_f; }
#define _GetInstrHit        n_instr_hit
/* DATA MACRO */
#define _AddDataFetch(v)   { n_data_fetch += (v); }
#define _GetDataFetch      n_data_fetch
#define _AddDataHit(v)     { n_data_hit += (v); }
#define _GetDataHit       n_data_hit

/* HISTORY MACRO */
#define _SetHistoryPos(v)   {history_pos = (v);}
#define _GetHistoryPos      history_pos

typedef int path_vect_t;
typedef struct c_sg_node_struct c_sg_node_t;
typedef struct path_struct path_t;
typedef struct i_fetch_list_struct i_fetch_list_t;
typedef struct child_list_struct child_list_t;
typedef struct layout_path_element_struct layout_path_element_t;
typedef struct task_list_struct task_list_t;
typedef struct header_path_struct header_path_t;
typedef struct path_cache_layout_struct path_cache_layout_t;
typedef struct layout_path_pair_struct layout_path_pair_t;
typedef struct confl_inter_struct confl_inter_t;


struct i_fetch_list_struct {
   /* int n_i_fetches;   number of instruction fetches within the node */
   int addr;         /* effective address */ 
   int width;        /* address width */
   i_fetch_list_t *next;
};

struct path_struct {
   int node;
   float hit_rate_after_path_exec; /* hit rate of the node after the execution
                                      of this path */
   float correct_percen;      /* correct percentage of presence in cache
                                 that must be considered for the current node;
                                 at the beginning it is set to 1.00
                                 (in alloc_node_path) , but then it can be 
                                 reduced by the add_correct_percen_path 
                                 due to conflicts with previous nodes 
                                 in the path */
   path_t *next;  /* pointer to the following node in the path */
};

struct header_path_struct {
   path_t *head;
   path_t *tail;
   path_vect_t *path_vect; /* it contains 1 if the current node 
                              belongs to the current header_path,
                              otherwise 0 */
   int n_i_fetches;        /* total number of instr. fetches */
};

struct child_list_struct {
   int child_id;
   float confl_instr;  /* confl instr value */
   child_list_t *child_list;
};
 
struct c_sg_node_struct {
   int type;         /* 0=BEGIN, 1=ASSIGN, 2=TEST, 3=END */
   int task_id;      /* task to which the node belongs */ 
   char *fuzzy_expr; /* fuzzy compilation info */
   int s_mem;        /* start address in memory */
   int e_mem;        /* end address in memory */
   int s_cache;      /* start set in cache */
   int e_cache;      /* end set in cache */
   float hit_rate;   /* current hit_rate of the node */
   int n_i_fetches;  /* n. of instruction references */
   i_fetch_list_t *i_fetch_list;  /* list of instruction references */
   i_fetch_list_t *last_i_fetch;
   /* Mauro */
   int n_d_fetches;  /* n. of data references */
   i_fetch_list_t *d_fetch_list;  /* list of data references */
   i_fetch_list_t *last_d_fetch;  /* same structure as instruction refs */
   /* fine Mauro */
   child_list_t *child_list;
   child_list_t *last_child;
};

struct task_list_struct {
   int num_c_sg_nodes;
   int start_addr;    /* offset for loading in memory */
   c_sg_node_t *sgraph;
};

struct path_cache_layout_struct {
   int start;
   int end;
   float weigth;  /* in case of associativity > 1 */
   path_cache_layout_t *next_segment;
};

struct layout_path_pair_struct {
   int task_id;
   int path_id;
   path_cache_layout_t *layout;
   header_path_t *path;
};

struct confl_inter_struct {
   int task1;
   int path1;
   int task2;
   int path2;
   float confl;
};

#ifdef UNIX
extern FILE * fptraces;
#endif /* UNIX */
extern int history_pos;
extern int n_instr_fetch;
extern float n_instr_hit;
extern int n_data_hit;
extern int n_instr_fetch_path;
extern float n_instr_hit_path;

extern task_list_t task[];
extern int cache_size;
extern int line_size;
extern int associativity;
extern int history_length;
extern float threshold;
extern int csimulstyle;

EXTERN int cache_path_lookup( int, int, c_sg_node_t *, header_path_t **, 
                              path_cache_layout_t **, int, int, int );
EXTERN void cache( header_path_t *, c_sg_node_t *, int, int, path_cache_layout_t **, int, int, int, int, FILE * );
EXTERN void cache_plot_hit_rate( int, c_sg_node_t *, header_path_t *, 
                                 int, int, int, int );
EXTERN void cache_layout( int, int, path_cache_layout_t *, header_path_t *, c_sg_node_t *, int, int, int, int );
EXTERN void compute_abstract_cache_state( int, int, c_sg_node_t *, 
                              header_path_t **, path_cache_layout_t **, 
                              int, int, int, int, FILE * );

EXTERN void print_traces( c_sg_node_t *, int, int, int, int, FILE * );

EXTERN void reset_cache( );
EXTERN void init_cache( int, int, int, int, float, int, int, int );

EXTERN task_list_t * get_task( int );
EXTERN int get_task_num_c_sg_nodes( task_list_t * );
EXTERN void set_task_num_c_sg_nodes( task_list_t *, int );
EXTERN int get_task_start_addr( task_list_t * );
EXTERN void set_task_start_addr( task_list_t *, int );
EXTERN c_sg_node_t * get_task_sgraph( task_list_t * );
EXTERN void set_task_sgraph( task_list_t *, c_sg_node_t *);

EXTERN void build_sgraph( int, int, int, char * );
EXTERN c_sg_node_t * create_sgraph( int, int );
EXTERN void alloc_sgraph( int , int );

EXTERN c_sg_node_t * get_node_sgraph( c_sg_node_t *, int );
EXTERN int get_task_id( c_sg_node_t * );
EXTERN void set_task_id( c_sg_node_t *, int );
EXTERN char * get_fuzzy_expr( const c_sg_node_t * );
EXTERN void set_fuzzy_expr( c_sg_node_t *, char * ); 
EXTERN int get_s_mem( const c_sg_node_t * );
EXTERN void set_s_mem( c_sg_node_t *, int );
EXTERN int get_e_mem( c_sg_node_t * );
EXTERN void set_e_mem( c_sg_node_t *, int );
EXTERN int get_s_cache( const c_sg_node_t * );
EXTERN void set_s_cache( c_sg_node_t *, int );
EXTERN int get_e_cache( const c_sg_node_t * );
EXTERN void set_e_cache( c_sg_node_t *, int );
EXTERN float get_hit_rate( c_sg_node_t * );
EXTERN void set_hit_rate( c_sg_node_t *, float );
EXTERN int get_n_i_fetches( c_sg_node_t * ); 
EXTERN void set_n_i_fetches( c_sg_node_t *, int );

EXTERN task_list_t * get_task( int );
EXTERN int get_task_num_c_sg_nodes( task_list_t * );
EXTERN void set_task_num_c_sg_nodes( task_list_t *, int );
EXTERN int get_task_start_addr( task_list_t * );
EXTERN void set_task_start_addr( task_list_t *, int );
EXTERN c_sg_node_t * get_task_sgraph( task_list_t * );
EXTERN void set_task_sgraph( task_list_t *, c_sg_node_t *);

EXTERN i_fetch_list_t * alloc_i_fetch_list( void );
EXTERN void del_i_fetch( i_fetch_list_t * );
EXTERN void free_i_fetch_list( c_sg_node_t *, int );
EXTERN i_fetch_list_t * get_i_fetch_list(c_sg_node_t *);
EXTERN i_fetch_list_t * get_d_fetch_list(c_sg_node_t *);
EXTERN void set_i_fetch_list(c_sg_node_t *, i_fetch_list_t *);

EXTERN void add_i_fetch(c_sg_node_t *, int , int );
EXTERN void add_d_fetch(c_sg_node_t *, int , int );
EXTERN int get_i_fetch_addr( i_fetch_list_t * );
EXTERN void set_i_fetch_addr( i_fetch_list_t *, int );
EXTERN i_fetch_list_t * create_i_fetch( int );
EXTERN i_fetch_list_t *  get_last_i_fetch( c_sg_node_t * );
EXTERN void set_last_i_fetch( c_sg_node_t *, i_fetch_list_t * );
EXTERN void set_next_i_fetch( i_fetch_list_t *, i_fetch_list_t * );
EXTERN i_fetch_list_t * get_next_i_fetch( i_fetch_list_t * );

EXTERN child_list_t * alloc_child_list( void );
EXTERN void free_child_list( child_list_t * );
EXTERN child_list_t * create_child( int );
EXTERN void set_child_list( c_sg_node_t *, child_list_t * );
EXTERN child_list_t * get_child_list( c_sg_node_t * );
EXTERN void set_child_id( child_list_t *, int );
EXTERN int get_child_id( child_list_t * );
EXTERN child_list_t *  get_last_child( c_sg_node_t * );
EXTERN void set_last_child( c_sg_node_t *, child_list_t *);
EXTERN void set_cache_confl_instr( c_sg_node_t *, int , int , float );
EXTERN float get_child_confl_instr( child_list_t * );
EXTERN void add_child( c_sg_node_t *, int, int );
EXTERN void set_next_child( child_list_t *, child_list_t * );
EXTERN child_list_t * get_next_child( child_list_t * );

EXTERN void set_addr_cache ( c_sg_node_t * );
EXTERN void set_addr_mem ( int, c_sg_node_t *, int, int, int );
EXTERN void update_addresses( c_sg_node_t *, int, int );
EXTERN float calc_confl( int, int, int, int );
EXTERN float calc_confl_instr( c_sg_node_t *, int, int );
EXTERN void set_confl_instr( c_sg_node_t *, int, int );

EXTERN int get_start_cache_segment( path_cache_layout_t * );
EXTERN void set_start_cache_segment( path_cache_layout_t *, int );
EXTERN int get_end_cache_segment( path_cache_layout_t *);
EXTERN void set_end_cache_segment( path_cache_layout_t *, int );
EXTERN path_cache_layout_t * get_next_cache_segment( path_cache_layout_t * );
EXTERN void set_next_cache_segment( path_cache_layout_t *, path_cache_layout_t * );
EXTERN path_cache_layout_t * create_path_cache_layout( c_sg_node_t * );
EXTERN void free_cache_segment( path_cache_layout_t * );
EXTERN void free_cache_layout( path_cache_layout_t * );
EXTERN path_cache_layout_t * update_cache_segment( path_cache_layout_t *, path_cache_layout_t * );
EXTERN path_cache_layout_t * add_node_to_layout( path_cache_layout_t *, c_sg_node_t *, int ); 
EXTERN path_cache_layout_t * update_cache_layout( path_cache_layout_t *, path_cache_layout_t * ); 
EXTERN void print_path_cache_layout( path_cache_layout_t * );

EXTERN int make_key( int, int  );
EXTERN layout_path_pair_t * lookup_layout_table( int, int );
EXTERN void insert_in_layout_table( layout_path_pair_t * );

EXTERN int get_node_path( path_t * );
EXTERN void set_node_path( path_t *, int );
EXTERN path_t * get_next_node_path( path_t * );
EXTERN void set_next_node_path( path_t *, path_t * );

EXTERN float get_hit_rate_after_path_exec( path_t * );
EXTERN void set_hit_rate_after_path_exec( path_t *, float );

EXTERN float get_correct_percen_path( path_t * );
EXTERN void set_correct_percen_path( path_t *, float );
EXTERN void add_correct_percen_path( path_t *, float );

EXTERN path_vect_t *alloc_path_vect( int );
EXTERN void free_path_vect( path_vect_t * ); 
EXTERN path_vect_t get_path_vect_pos( header_path_t *, int );
EXTERN void set_path_vect_pos( header_path_t *, int );
EXTERN path_vect_t * get_path_vect( header_path_t * );
EXTERN void set_path_vect( header_path_t *, path_vect_t * );

EXTERN header_path_t * alloc_header_path( int );
EXTERN path_t * get_head_path( header_path_t * );
EXTERN void set_head_path( header_path_t *, path_t * );
EXTERN path_t * get_tail_path( header_path_t * );
EXTERN void set_tail_path( header_path_t *, path_t * );
EXTERN int get_n_i_fetches_path( header_path_t * );
EXTERN void set_n_i_fetches_path( header_path_t *, int );
EXTERN void add_n_i_fetches_path( header_path_t *, int );
EXTERN path_t * alloc_node_path( void );
EXTERN void free_node_path( path_t * );
EXTERN void free_path( header_path_t * );
EXTERN path_t * create_node_path( int );
EXTERN void add_node_in_path( header_path_t *, int, int, int );

/* HISTORY CACHE */
EXTERN void alloc_history_cache( void );
EXTERN void free_history_cache( void );
EXTERN layout_path_pair_t * get_history_elem( int );
EXTERN void set_history_elem( layout_path_pair_t *, int );
EXTERN int insert_in_history( layout_path_pair_t *, int );
EXTERN layout_path_pair_t * lookup_history( int, int );
EXTERN void print_history( void );

EXTERN int get_task_id_layout_path_pair( layout_path_pair_t * );
EXTERN void set_task_id_layout_path_pair( layout_path_pair_t *, int );
EXTERN int get_path_id_layout_path_pair( layout_path_pair_t * );
EXTERN void set_path_id_layout_path_pair( layout_path_pair_t *, int );
EXTERN path_cache_layout_t *get_layout_layout_path_pair( layout_path_pair_t * );
EXTERN void set_layout_layout_path_pair( layout_path_pair_t *, path_cache_layout_t * );
EXTERN header_path_t *get_path_layout_path_pair( layout_path_pair_t * );
EXTERN void set_path_layout_path_pair( layout_path_pair_t *, header_path_t * );
EXTERN layout_path_pair_t * create_layout_path_pair( int, int, path_cache_layout_t *, header_path_t * );
EXTERN void free_layout_path_pair( layout_path_pair_t * );

EXTERN void compute_intra_conflicts( header_path_t *, c_sg_node_t *, int );
EXTERN float update_intra_conflicts( header_path_t *, c_sg_node_t * );
EXTERN void record_path_hit_rate( header_path_t *, c_sg_node_t * );
EXTERN void print_path_hit_rate( header_path_t *, c_sg_node_t * );

EXTERN float compute_inter_conflicts( path_cache_layout_t *,path_cache_layout_t * );
EXTERN void update_inter_conflicts( layout_path_pair_t *, int );
EXTERN void propagate_inter_in_intra_conflicts( header_path_t *, c_sg_node_t *, float );
EXTERN path_cache_layout_t *create_diff_path_cache_layout( header_path_t *, header_path_t *, int, header_path_t * ); 

EXTERN int get_cache_size( void );
EXTERN void set_cache_size( int );
EXTERN int get_line_size( void );
EXTERN void set_line_size( int );
EXTERN int get_associativity( void );
EXTERN void set_associativity( int );
EXTERN int get_history_length( void );
EXTERN void set_history_length( int );
EXTERN float get_threshold( void );
EXTERN void set_threshold( float );
EXTERN int get_csimulstyle( void );
EXTERN void set_csimulstyle( int );
EXTERN int get_i_fetchsize( void );
EXTERN void set_i_fetchsize( int );
EXTERN int get_miss_penalty( void );
EXTERN void set_miss_penalty( int );

EXTERN int get_task_id1_confl( confl_inter_t * );
EXTERN void set_task_id1_confl( confl_inter_t *, int );
EXTERN int get_task_id2_confl( confl_inter_t * );
EXTERN void set_task_id2_confl( confl_inter_t *, int );
EXTERN int get_path_id1_confl( confl_inter_t * );
EXTERN void set_path_id1_confl( confl_inter_t *, int );
EXTERN int get_path_id2_confl( confl_inter_t * );
EXTERN void set_path_id2_confl( confl_inter_t *, int );
EXTERN int get_confl_confl( confl_inter_t * );
EXTERN void set_confl_confl( confl_inter_t *, float );
EXTERN confl_inter_t * create_inter_confl( int, int, int, int, float );
EXTERN void free_inter_confl( confl_inter_t * );
EXTERN int make_confl_key( int, int, int, int );
EXTERN confl_inter_t * get_confl_from_table( int );
EXTERN void set_confl_in_table( int, confl_inter_t * );
EXTERN void clean_confl_inter_table( void );

EXTERN void insert_in_confl_inter_table( int, int, int, int, float );
EXTERN float lookup_confl_inter_table( int, int, int, int );


/* FULL CACHE SIMULATION */

EXTERN void print_cache( int * );
EXTERN int put_in_cache ( int *, i_fetch_list_t * );

/* getcmdargs.c */
#ifdef UNIX
EXTERN int getcmdargs( int, char ** );
#endif

EXTERN void PrintTraces( FILE *, int );

EXTERN void clean_hit_rate( c_sg_node_t *, int );
