-- ----------------------------------------------------------------------------
--   Title      : NUMERIC_BIT arithmetic package
--              : Rev. 1.8 (Dec. 14 1994)
--              :
--   Library    : This package shall be compiled into the WORK library
--              :
--   Developers : IEEE DASC Synthesis Working Group, PAR 1076.3 
--		: Adapted by Bassam Tabbara, tbassam@ic.eecs.berkeley.edu
--		: University of California, Berkeley
--		: for use with the VHDL Co-simulation Package in Polis
--              :
--   Purpose    : This package defines numeric types and arithmetic functions 
--              : for use with synthesis tools. Two numeric types are defined: 
--              :     --> UNSIGNED : represents UNSIGNED number in vector form
--              :     --> SIGNED   : represents a SIGNED number in vector form
--              : The base element type is type BIT.
--              : The LEFTmost bit is treated as the most significant bit.
--              : Signed vectors are represented in two's complement form.
--              : This package contains overloaded arithmetic operators on 
--              : the SIGNED and UNSIGNED types. The package also contains
--              : useful type conversions functions, clock detection 
--              : functions, and other utility functions.
--              :
--              : If any argument to a function is a null array, a null array is
--              : returned (exceptions, if any, are noted individually).
--              :
--   Note       : No declarations or definitions shall be included in, or
--              : excluded from, this package. The package declaration declares
--              : the functions that can be used by a user. The package body
--              : shall be considered the formal definition of the semantics of
--              : this package. Tool developers may choose to implement the 
--              : package body in the most efficient manner available to them.
--              :
-- ----------------------------------------------------------------------------

Package NUMERIC_BIT is
 
  --===========================================================================
  -- Numeric array type definitions
  --===========================================================================
 
  type UNSIGNED is array ( NATURAL range <> ) of BIT;
  type SIGNED is array ( NATURAL range <> ) of BIT;
 
  --===========================================================================
  -- Arithmetic Operators:
  --===========================================================================
 
     -- Id: A.1
  function "abs" ( x : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(X'LENGTH-1 downto 0).
     -- Result: Returns the absolute value of a SIGNED vector X.
 
     -- Id: A.2
  function "-" ( ARG : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(ARG'LENGTH-1 downto 0).
     -- Result: Returns the value of the unary minus operation on a
     --         SIGNED vector ARG.
 
  --============================================================================

     -- Id: A.3
  function "+" ( L,R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED(MAX(L'LENGTH, R'LENGTH)-1 downto 0).
     -- Result: Adds two UNSIGNED vectors that may be of different lengths.
 
     -- Id: A.4
  function "+" ( L,R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(MAX(L'LENGTH, R'LENGTH)-1 downto 0).
     -- Result: Adds two SIGNED vectors that may be of different lengths.
 
     -- Id: A.5
  function "+" ( L : UNSIGNED; R : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED(L'LENGTH-1 downto 0).
     -- Result: Adds an UNSIGNED vector, L, with a non-negative integer, R.
 
     -- Id: A.6
  function "+" ( L : NATURAL; R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED(R'LENGTH-1 downto 0).
     -- Result: Adds a non-negative integer, L, with an UNSIGNED vector, R.
 
     -- Id: A.7
  function "+" ( L : INTEGER; R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED (R'LENGTH-1 downto 0).
     -- Result: Adds an integer, L (may be positive or negative), to a SIGNED
     --         vector, R.
 
     -- Id: A.8
  function "+" ( L : SIGNED; R : INTEGER ) return SIGNED;
     -- Result subtype: SIGNED (L'LENGTH-1 downto 0).
     -- Result: Adds a SIGNED vector, L, to an integer, R.
 
  --============================================================================
 
     -- Id: A.9
  function "-" ( L,R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED(MAX(L'LENGTH, R'LENGTH)-1 downto 0).
     -- Result: Subtracts two UNSIGNED vectors that may be of different lengths.
 
     -- Id: A.10
  function "-" ( L,R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(MAX(L'LENGTH, R'LENGTH)-1 downto 0).
     -- Result: Subtracts a SIGNED vector, R, from another SIGNED vector, L,
     --         that may possibly be of different lengths.
 
     -- Id: A.11
  function "-" ( L : UNSIGNED; R : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED (L'LENGTH-1 downto 0).
     -- Result: Subtracts a non-negative integer, R, from an UNSIGNED vector, L.
 
     -- Id: A.12
  function "-" ( L : NATURAL; R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED(R'LENGTH-1 downto 0).
     -- Result: Subtracts an UNSIGNED vector, R, from a non-negative integer, L.
 
     -- Id: A.13
  function "-" ( L : SIGNED; R : INTEGER ) return SIGNED;
     -- Result subtype: SIGNED (L'LENGTH-1 downto 0).
     -- Result: Subtracts an integer, R, from a SIGNED vector, L.
 
     -- Id: A.14
  function "-" ( L : INTEGER; R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(R'LENGTH-1 downto 0).
     -- Result: Subtracts a SIGNED vector, R, from an integer, L.
 
  --============================================================================
 
     -- Id: A.15
  function "*" ( L,R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED((L'length+R'length-1) downto 0).
     -- Result: Performs the multiplication operation on two UNSIGNED vectors
     --         that may possibly be of different lengths.
 
     -- Id: A.16
  function "*" ( L,R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED((L'length+R'length-1) downto 0)
     -- Result: Multiplies two SIGNED vectors that may possibly be of
     --         different lengths.
 
     -- Id: A.17
  function "*" ( L : UNSIGNED; R : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED((L'length+L'length-1) downto 0).
     -- Result: Multiplies an UNSIGNED vector, L, with a non-negative 
     --         integer, R. R is converted to an UNSIGNED vector of 
     --         size L'length before multiplication.
 
     -- Id: A.18
  function "*" ( L : NATURAL; R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED((R'length+R'length-1) downto 0).
     -- Result: Multiplies an UNSIGNED vector, R, with a non-negative 
     --         integer, L. L is converted to an UNSIGNED vector of 
     --         size R'length before multiplication.
 
     -- Id: A.19
  function "*" ( L : SIGNED; R : INTEGER ) return SIGNED;
     -- Result subtype: SIGNED((L'length+L'length-1) downto 0)
     -- Result: Multiplies a SIGNED vector, L, with an integer, R. R is
     --         converted to a SIGNED vector of size L'length before 
     --         multiplication.
 
     -- Id: A.20
  function "*" ( L : INTEGER; R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED((R'length+R'length-1) downto 0)
     -- Result: Multiplies a SIGNED vector, R, with an integer, L. L is
     --         converted to a SIGNED vector of size R'length before 
     --         multiplication.
 
  --============================================================================
  --
  --  NOTE: If second argument is zero for "/" operator, a severity level
  --        of ERROR is issued.
 
     -- Id: A.21
  function "/" ( L,R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED (L'LENGTH-1 downto 0)
     -- Result: Divides an UNSIGNED vector, L, by another UNSIGNED vector, R.
 
     -- Id: A.22
  function "/" ( L,R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED (L'LENGTH-1 downto 0)
     -- Result: Divides an SIGNED vector, L, by another SIGNED vector, R.
 
     -- Id: A.23
  function "/" ( L : UNSIGNED; R : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED (L'LENGTH-1 downto 0)
     -- Result: Divides an UNSIGNED vector, L, by a non-negative integer, R.
     --         If NO_OF_BITS(R) > L'LENGTH, then R is truncated to L'LENGTH.
 
     -- Id: A.24
  function "/" ( L : NATURAL; R : UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED (R'LENGTH-1 downto 0)
     -- Result: Divides a non-negative integer, L, by an UNSIGNED vector, R.
     --         If NO_OF_BITS(L) > R'LENGTH, then L is truncated to R'LENGTH.
 
     -- Id: A.25
  function "/" ( L : SIGNED; R : INTEGER ) return SIGNED;
     -- Result subtype: SIGNED (L'LENGTH-1 downto 0)
     -- Result: Divides a SIGNED vector, L, by an integer, R.
     --         If NO_OF_BITS(R) > L'LENGTH, then R is truncated to L'LENGTH.
 
     -- Id: A.26
  function "/" ( L : INTEGER; R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED (R'LENGTH-1 downto 0)
     -- Result: Divides an integer, L, by a SIGNED vector, R.
     --         If NO_OF_BITS(L) > R'LENGTH, then L is truncated to R'LENGTH.
 
  --============================================================================
  --
  --  NOTE: If second argument is zero for "rem" operator, a severity level
  --        of ERROR is issued.
 
     -- Id: A.27
  function "rem" ( L,R: UNSIGNED)return UNSIGNED;
     -- Result subtype: UNSIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L rem R" where L and R are UNSIGNED vectors.
 
     -- Id: A.28
  function "rem" ( L,R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L rem R" where L and R are SIGNED vectors.
 
     -- Id: A.29
  function "rem" ( L : UNSIGNED; R: NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L rem R" where L is an UNSIGNED vector and R is a 
     --         non-negative integer.
     --         If NO_OF_BITS(R) > L'LENGTH, then R is truncated to L'LENGTH.
 
     -- Id: A.30
  function "rem" ( L : NATURAL; R: UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED(R'LENGTH-1 downto 0)
     -- Result: Computes "L rem R" where R is an UNSIGNED vector and L is a 
     --         non-negative integer.
     --         If NO_OF_BITS(L) > R'LENGTH, then L is truncated to R'LENGTH.
 
     -- Id: A.31
  function "rem" ( L : SIGNED; R : INTEGER ) return SIGNED;
     -- Result subtype: SIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L rem R" where L is SIGNED vector and R is an integer.
     --         If NO_OF_BITS(R) > L'LENGTH, then R is truncated to L'LENGTH.
 
     -- Id: A.32
  function "rem" ( L : INTEGER; R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(R'LENGTH-1 downto 0)
     -- Result: Computes "L rem R" where R is SIGNED vector and L is an integer.
     --         If NO_OF_BITS(L) > R'LENGTH, then L is truncated to R'LENGTH.
 
  --============================================================================
  --
  --  NOTE: If second argument is zero for "mod" operator, a severity level
  --        of ERROR is issued.
 
     -- Id: A.33
  function "mod" ( L,R: UNSIGNED)return UNSIGNED;
     -- Result subtype: UNSIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L mod R" where L and R are UNSIGNED vectors.
 
     -- Id: A.34
  function "mod" ( L,R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L mod R" where L and R are SIGNED vectors.
 
     -- Id: A.35
  function "mod" ( L : UNSIGNED; R: NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L mod R" where L is an UNSIGNED vector and R 
     --         is a non-negative integer.
     --         If NO_OF_BITS(R) > L'LENGTH, then R is truncated to L'LENGTH.
 
     -- Id: A.36
  function "mod" ( L : NATURAL; R: UNSIGNED) return UNSIGNED;
     -- Result subtype: UNSIGNED(R'LENGTH-1 downto 0)
     -- Result: Computes "L mod R" where R is an UNSIGNED vector and L 
     --         is a non-negative integer.
     --         If NO_OF_BITS(L) > R'LENGTH, then L is truncated to R'LENGTH.
 
     -- Id: A.37
  function "mod" ( L : SIGNED; R : INTEGER ) return SIGNED;
     -- Result subtype: SIGNED(L'LENGTH-1 downto 0)
     -- Result: Computes "L mod R" where L is a SIGNED vector and
     --         R is an integer.
     --         If NO_OF_BITS(R) > L'LENGTH, then R is truncated to L'LENGTH.
 
     -- Id: A.38
  function "mod" ( L : INTEGER; R : SIGNED) return SIGNED;
     -- Result subtype: SIGNED(R'LENGTH-1 downto 0)
     -- Result: Computes "L mod R" where L is an integer and
     --         R is a SIGNED vector.
     --         If NO_OF_BITS(L) > R'LENGTH, then L is truncated to R'LENGTH.
 
  --============================================================================
  -- Comparison Operators
  --============================================================================
 
     -- Id: C.1
  function ">"  ( L,R : UNSIGNED) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L > R" where L and R are UNSIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.2
  function ">"  ( L,R : SIGNED)     return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L > R" where L and R are SIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.3
  function ">"  ( L : NATURAL; R : UNSIGNED)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L > R" where L is a non-negative integer and
     --         R is an UNSIGNED vector.
 
     -- Id: C.4
  function ">"  ( L : INTEGER; R : SIGNED)    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L > R" where L is a integer and
     --         R is a SIGNED vector.
 
     -- Id: C.5
  function ">"  ( L : UNSIGNED; R : NATURAL)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L > R" where L is an UNSIGNED vector and
     --         R is a non-negative integer.

     -- Id: C.6
  function ">"  ( L : SIGNED; R : INTEGER )    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L > R" where L is a SIGNED vector and
     --         R is a integer.
 
  --============================================================================
 
     -- Id: C.7
  function "<"  ( L,R : UNSIGNED) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L < R" where L and R are UNSIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.8
  function "<"  ( L,R : SIGNED)     return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L < R" where L and R are SIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.9
  function "<"  ( L : NATURAL; R : UNSIGNED)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L < R" where L is a non-negative integer and
     --         R is an UNSIGNED vector.
 
     -- Id: C.10
  function "<"  ( L : INTEGER; R : SIGNED)    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L < R" where L is an integer and
     --         R is a SIGNED vector.
 
     -- Id: C.11
  function "<"  ( L : UNSIGNED; R : NATURAL)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L < R" where L is an UNSIGNED vector and
     --         R is a non-negative integer.
 
     -- Id: C.12
  function "<"  ( L : SIGNED; R : INTEGER )    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L < R" where L is a SIGNED vector and
     --         R is an integer.
 
  --============================================================================
 
     -- Id: C.13
  function "<=" ( L,R : UNSIGNED) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L <= R" where L and R are UNSIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.14
  function "<=" ( L,R : SIGNED)     return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L <= R" where L and R are SIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.15
  function "<=" ( L : NATURAL; R : UNSIGNED)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L <= R" where L is a non-negative integer and
     --         R is an UNSIGNED vector.
 
     -- Id: C.16
  function "<=" ( L : INTEGER; R : SIGNED)    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L <= R" where L is an integer and
     --         R is a SIGNED vector.
 
     -- Id: C.17
  function "<=" ( L : UNSIGNED; R : NATURAL)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L <= R" where L is an UNSIGNED vector and
     --         R is a non-negative integer.

     -- Id: C.18
  function "<=" ( L : SIGNED; R : INTEGER )    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L <= R" where L is a SIGNED vector and
     --         R is an integer.
 
  --============================================================================
 
     -- Id: C.19
  function ">=" ( L,R : UNSIGNED) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L >= R" where L and R are UNSIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.20
  function ">=" ( L,R : SIGNED)     return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L >= R" where L and R are SIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.21
  function ">=" ( L : NATURAL; R : UNSIGNED)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L >= R" where L is a non-negative integer and
     --         R is an UNSIGNED vector.
 
     -- Id: C.22
  function ">=" ( L : INTEGER; R : SIGNED)    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L >= R" where L is an integer and
     --         R is a SIGNED vector.
 
     -- Id: C.23
  function ">=" ( L : UNSIGNED; R : NATURAL)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L >= R" where L is an UNSIGNED vector and
     --         R is a non-negative integer.
 
     -- Id: C.24
  function ">=" ( L : SIGNED; R : INTEGER )    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L >= R" where L is a SIGNED vector and
     --         R is an integer.

  --============================================================================
 
     -- Id: C.25
  function "=" ( L,R : UNSIGNED) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L = R" where L and R are UNSIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.26
  function "=" ( L,R : SIGNED)     return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L = R" where L and R are SIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.27
  function "=" ( L : NATURAL; R : UNSIGNED)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L = R" where L is a non-negative integer and
     --         R is an UNSIGNED vector.
 
     -- Id: C.28
  function "=" ( L : INTEGER; R : SIGNED)    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L = R" where L is an integer and
     --         R is a SIGNED vector.
 
     -- Id: C.29
  function "=" ( L : UNSIGNED; R : NATURAL)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L = R" where L is an UNSIGNED vector and
     --         R is a non-negative integer.
 
     -- Id: C.30
  function "=" ( L : SIGNED; R : INTEGER )    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L = R" where L is a SIGNED vector and
     --         R is an integer.
 
  --============================================================================
 
     -- Id: C.31
  function "/=" ( L,R : UNSIGNED) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L /= R" where L and R are UNSIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.32
  function "/=" ( L,R : SIGNED)     return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L /= R" where L and R are SIGNED vectors possibly
     --         of different lengths.
 
     -- Id: C.33
  function "/=" ( L : NATURAL; R : UNSIGNED)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L /= R" where L is a non-negative integer and
     --         R is an UNSIGNED vector.
 
     -- Id: C.34
  function "/=" ( L : INTEGER; R : SIGNED)    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L /= R" where L is an integer and
     --         R is a SIGNED vector.
 
     -- Id: C.35
  function "/=" ( L : UNSIGNED; R : NATURAL)  return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L /= R" where L is an UNSIGNED vector and
     --         R is a non-negative integer.
 
     -- Id: C.36
  function "/=" ( L : SIGNED; R : INTEGER )    return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Computes "L /= R" where L is a SIGNED vector and
     --         R is an integer.
 
  --============================================================================
  -- Shift and Rotate Functions
  --============================================================================
 
     -- Id: S.1
  function shift_LEFT  (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a shift-LEFT on an UNSIGNED vector COUNT times.
     --         The vacated positions are filled with Bit '0'.
     -- The COUNT LEFTmost bits are lost.
 
     -- Id: S.2
  function shift_RIGHT (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a shift-RIGHT on an UNSIGNED vector COUNT times.
     --         The vacated positions are filled with Bit '0'.
     --         The COUNT RIGHTmost bits are lost.
 
     -- Id: S.3
  function shift_LEFT  (  ARG  : SIGNED; count : NATURAL) return SIGNED;
     -- Result subtype: SIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a shift-LEFT on a SIGNED vector COUNT times.
     --         The vacated positions are filled with Bit '0'. 
     --         The COUNT LEFTmost bits, except ARG'LEFT, are lost.
 
     -- Id: S.4
  function shift_RIGHT (  ARG  : SIGNED; count : NATURAL) return SIGNED;
     -- Result subtype: SIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a shift-RIGHT on a SIGNED vector COUNT times.
     --         The vacated positions are filled with the LEFTmost bit,ARG'LEFT.
     --         The COUNT RIGHTmost bits are lost.
 
  --============================================================================
 
  --============================================================================
 
     -- Id: S.5
  function ROTATE_LEFT  (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a ROTATE_LEFT of an UNSIGNED vector COUNT times.

     -- Id: S.6
  function ROTATE_RIGHT (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a ROTATE_RIGHT of an UNSIGNED vector COUNT times.
 
     -- Id: S.7
  function ROTATE_LEFT  (  ARG  : SIGNED; count : NATURAL) return SIGNED;
     -- Result subtype: SIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a logical rotate-LEFT of a SIGNED 
     --         vector COUNT times.
 
     -- Id: S.8
  function ROTATE_RIGHT (  ARG  : SIGNED; count : NATURAL) return SIGNED;
     -- Result subtype: SIGNED (ARG'LENGTH-1 downto 0)
     -- Result: Performs a logical rotate-RIGHT of a SIGNED 
     --         vector COUNT times.
 
  --============================================================================
  --  RESIZE Functions
  --============================================================================

     -- Id: R.1
  function RESIZE ( ARG : SIGNED; new_size : NATURAL) return SIGNED;
     -- Result subtype: SIGNED(new_size-1 downto 0)
     -- Result: Resizes the SIGNED vector ARG to the specified size. 
     --         To create a larger vector, the new [LEFTmost] bit positions
     --         are filled with the sign bit (ARG'LEFT). When truncating,
     --         the sign bit is retained along with the RIGHTmost part.
 
     -- Id: R.2
  function RESIZE ( ARG : UNSIGNED; new_size : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED(new_size-1 downto 0)
     -- Result: Resizes the UNSIGNED vector ARG to the specified size. 
     --         To create a larger vector, the new [LEFTmost] bit positions
     --         are filled with '0'. When truncating, the LEFTmost bits 
     --         are dropped.

 
  --============================================================================
  -- Conversion Functions
  --============================================================================
 
     -- Id: D.1
  function TO_INTEGER ( ARG : UNSIGNED) return NATURAL;
     -- Result subtype: NATURAL. Value cannot be negative since parameter is an
     --                 UNSIGNED vector.
     -- Result: Converts the UNSIGNED vector to an integer.
 
     -- Id: D.2
  function TO_INTEGER ( ARG : SIGNED) return INTEGER;
     -- Result subtype: INTEGER
     -- Result: Converts a SIGNED vector to an integer.
 
     -- Id: D.3
  function TO_UNSIGNED (  ARG,SIZE : NATURAL) return UNSIGNED;
     -- Result subtype: UNSIGNED (SIZE-1 downto 0)
     -- Result: Converts a non-negative integer to an UNSIGNED vector with
     --         the specified size.
 
     -- Id: D.4
  function TO_SIGNED (  ARG : INTEGER; SIZE : NATURAL) return SIGNED;
     -- Result subtype: SIGNED (SIZE-1 downto 0)
     -- Result: Converts an integer to a SIGNED vector of the specified size.
 
     -- Id: D.5
  function TO_UNSIGNED (  ARG : BIT_VECTOR) return UNSIGNED;
     -- Result subtype: UNSIGNED(ARG'RANGE)
     -- Result: Converts BIT_VECTOR to UNSIGNED.

     -- Id: D.6
  function TO_SIGNED (  ARG : BIT_VECTOR) return SIGNED;
     -- Result subtype: SIGNED(ARG'RANGE)
     -- Result: Converts BIT_VECTOR to SIGNED.

     -- Id: D.7
  function TO_BITVECTOR (  ARG : UNSIGNED) return BIT_VECTOR;
     -- Result subtype: BIT_VECTOR(ARG'RANGE)
     -- Result: Converts UNSIGNED to BIT_VECTOR.

     -- Id: D.8
  function TO_BITVECTOR (  ARG : SIGNED) return BIT_VECTOR;
     -- Result subtype: BIT_VECTOR(ARG'RANGE)
     -- Result: Converts SIGNED to BIT_VECTOR.

  --============================================================================
  -- Edge Detection Functions
  --============================================================================
 
     -- Id: E.1
   function RISING_EDGE  (SIGNAL s : BIT) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Returns TRUE if an event is detected on signal S and the
     --         value changed from a '0' to a '1'.
 
     -- Id: E.2
   function FALLING_EDGE (SIGNAL s : BIT) return BOOLEAN;
     -- Result subtype: BOOLEAN
     -- Result: Returns TRUE if an event is detected on signal S and the
     --         value changed from a '1' to a '0'.
 
end NUMERIC_BIT;


--=============================================================================
--=======================    Package    Body    ===============================
--=============================================================================


Package body NUMERIC_BIT is


   -- null range array constants
 
   constant nau : UNSIGNED (0 downto 1) := (others => '0');
   constant nas : SIGNED (0 downto 1) := (others => '0');
 
   -- implementation controls
   
   constant NO_WARNING : boolean := FALSE;  -- default to emit warnings

--=========================Local Subprograms=================================

function MAX(LEFT, RIGHT : integer) return integer is
begin
  if LEFT > RIGHT then return LEFT;
  else return RIGHT;
    end if;
  end MAX;

function MIN(LEFT, RIGHT : integer) return integer is
begin
  if LEFT < RIGHT then return LEFT;
    else return RIGHT;
    end if;
  end MIN;

    function Signed_Num_Bits ( ARG: integer) return natural is
       variable nBits: natural;
       variable n: natural;
    begin
       if ARG>=0 then
	  n:=ARG;
       else
	  n:=-(ARG+1);
       end if;
       nBits:=1;
       while n>0 loop
	  nBits:=nBits+1;
	  n:= n / 2;
       end loop;
       return nBits;
    end Signed_Num_Bits;

    function UnSIGNED_Num_Bits (ARG: natural) return natural is
       variable nBits: natural;
       variable n: natural;
    begin
       n:=ARG;
       nBits:=1;
       while n>1 loop
	  nBits:=nBits+1;
	  n:= n / 2;
       end loop;
       return nBits;
    end UnSIGNED_Num_Bits;


------------------------------------------------------------------------
   -- this internal function computes the addition of two UNSIGNED
   -- with input carry
   -- * the two arguments are of the same length

function ADD_UNSIGNED ( L,R : UNSIGNED; C : BIT ) return UNSIGNED  is
  constant L_LEFT:integer := L'length-1;
  alias XL : UNSIGNED(L_LEFT downto 0) is L;
  alias XR : UNSIGNED(L_LEFT downto 0) is R;
  variable RESULT : UNSIGNED(L_LEFT downto 0);
  variable cbit : bit := C;
begin
   for i in 0 to L_LEFT loop
     RESULT(i) := Cbit xor XL(i) xor XR(i);
     Cbit := (Cbit and XL(i)) or (Cbit and XR(i)) or (XL(i) and XR(i));
   end loop;
   return RESULT;
end ADD_UNSIGNED;

   -- this internal function computes the addition of two SIGNED
   -- with input carry
   -- * the two arguments are of the same length

function ADD_SIGNED ( L,R : SIGNED; C : BIT ) return SIGNED  is
  constant L_LEFT:integer := L'length-1;
  alias XL : SIGNED(L_LEFT downto 0) is L;
  alias XR : SIGNED(L_LEFT downto 0) is R;
  variable RESULT : SIGNED(L_LEFT downto 0);
  variable Cbit : bit := C;
begin
   for i in 0 to L_LEFT loop
     RESULT(i) := Cbit xor XL(i) xor XR(i);
     Cbit := (Cbit and XL(i)) or (Cbit and XR(i)) or (XL(i) and XR(i));
   end loop;
   return RESULT;
end ADD_SIGNED;

------------------------------------------------------------------------

    -- this internal procedure computes UNSIGNED division
    -- giving the quotient and remainder.
    procedure divMod (num, XDENOM: UNSIGNED;
                      xquot, xremain: out UNSIGNED) is
       variable TEMP: UNSIGNED(num'length downto 0);
       variable quot: UNSIGNED(MAX(num'length,XDENOM'length)-1 downto 0);
       alias DENOM : UNSIGNED (XDENOM'length-1 downto 0) is XDENOM;
       variable TOPBIT: integer;
    begin
       TEMP:="0"&num;
       quot:= (others =>'0');
       TOPBIT:=-1;
       for j in DENOM'range loop
          if DENOM(j)='1' then
             TOPBIT:=j;
             exit;
          end if;
       end loop;
       assert TOPBIT>=0  report "DIV,MOD,or REM by zero"   severity error;

       for j in num'length-(TOPBIT+1) downto 0 loop
          if TEMP(TOPBIT+j+1 downto j) >= "0"&DENOM(TOPBIT downto 0) then
             TEMP(TOPBIT+j+1 downto j) := (TEMP(TOPBIT+j+1 downto j))
                                       -("0"&DENOM(TOPBIT downto 0));
             quot(j):='1';
          end if;
	  assert TEMP(TOPBIT+j+1)='0'
		report "internal error in the division algorithm"
		severity error;
       end loop;
       xquot:=quot(num'length-1 downto 0);
       xremain:=TEMP(num'length-1 downto 0);
    end;

-----------------Local Subprograms - shift/rotate ops-------------------------

function xsll ( ARG  : BIT_VECTOR; count : NATURAL) return BIT_VECTOR is
  constant ARG_l:integer := ARG'length-1;
  alias XARG : bit_vector(ARG_l downto 0) is ARG;
  variable result : bit_vector(ARG_l downto 0) := (others=>'0');
begin
  if count <= ARG_l then
    result(ARG_l downto count):=XARG(ARG_l-count downto 0);
    end if;
  return result;
end;

function xsrl ( ARG  : BIT_VECTOR; count : NATURAL) return BIT_VECTOR is
  constant ARG_l:integer := ARG'length-1;
  alias XARG : bit_vector(ARG_l downto 0) is ARG;
  variable result : bit_vector(ARG_l downto 0) := (others=>'0');
begin
  if count <= ARG_l then
    result(ARG_l-count downto 0):=XARG(ARG_l downto count);
    end if;
  return result;
end;

function xsra ( ARG  : BIT_VECTOR; count : NATURAL) return BIT_VECTOR is
  constant ARG_l:integer := ARG'length-1;
  alias XARG : bit_vector(ARG_l downto 0) is ARG;
  variable result : bit_vector(ARG_l downto 0);
  variable xcount : natural := count;
begin
  if ((ARG'length <= 1) or (xcount = 0)) then return ARG;
  else 
    if (xcount > ARG_l) then xcount := ARG_l; end if;
    result(ARG_l-xcount downto 0):=XARG(ARG_l downto xcount);
    result(ARG_l downto (ARG_l - xcount + 1)) := (others=>XARG(ARG_l));
    end if;
  return result;
end;

function xrol ( ARG  : BIT_VECTOR; count : NATURAL) return BIT_VECTOR is
  constant ARG_l:integer := ARG'length-1;
  alias XARG : bit_vector(ARG_l downto 0) is ARG;
  variable result : bit_vector(ARG_l downto 0) := XARG;
  variable countm : integer;
begin
  countm := count mod (ARG_l + 1);
  if countm /= 0 then
    result(ARG_l downto countm):=XARG(ARG_l-countm downto 0);
    result(countm-1 downto 0):=XARG(ARG_l downto ARG_l-countm+1);
    end if;
  return result;
end;

function xror ( ARG  : BIT_VECTOR; count : NATURAL) return BIT_VECTOR is
  constant ARG_l:integer := ARG'length-1;
  alias XARG : bit_vector(ARG_l downto 0) is ARG;
  variable result : bit_vector(ARG_l downto 0) := XARG;
  variable countm : integer;
begin
  countm := count mod (ARG_l + 1);
  if countm /= 0 then
    result(ARG_l-countm downto 0):=XARG(ARG_l downto countm);
    result(ARG_l downto ARG_l-countm+1):=XARG(countm-1 downto 0);
    end if;
  return result;
end;


-----------------Local Subprograms - Relational ops--------------------------

--
-- General "=" for UNSIGNED vectors, same length
--
function UNSIGNED_equal ( L,R : UNSIGNED) return BOOLEAN is
begin
  return BIT_VECTOR (L) = BIT_VECTOR (R) ;
  end UNSIGNED_equal ;

--
-- General "=" for SIGNED vectors, same length
--
function SIGNED_equal ( L,R : SIGNED) return BOOLEAN is
begin
  return BIT_VECTOR (L) = BIT_VECTOR (R) ;
  end SIGNED_equal ;



--
-- General "<" for UNSIGNED vectors, same length
--
function UNSIGNED_less ( L,R : UNSIGNED) return BOOLEAN is
begin
  return BIT_VECTOR (L) < BIT_VECTOR (R) ;
  end UNSIGNED_less ;

--
-- General "<" function for SIGNED vectors, same length 
--
function SIGNED_less ( L,R : SIGNED) return BOOLEAN is
  -- Need aliasses to assure index direction
  variable intern_l : SIGNED (0 to L'LENGTH-1);
  variable intern_r : SIGNED (0 to R'LENGTH-1);
begin
  intern_l:=l;
  intern_r:=r;
  intern_l(0):=not intern_l(0);
  intern_r(0):=not intern_r(0);
  return bit_vector (intern_L) < bit_vector (intern_R) ;
end SIGNED_less ; 

--
-- General "<=" function for UNSIGNED vectors, same length 
--
function UNSIGNED_less_or_equal ( L,R : UNSIGNED) return BOOLEAN is
begin
  return BIT_VECTOR (L) <= BIT_VECTOR (R) ;
  end UNSIGNED_less_or_equal ;

--
-- General "<=" function for SIGNED vectors, same length 
--
function SIGNED_less_or_equal ( L,R : SIGNED) return BOOLEAN is
  -- Need aliasses to assure index direction
  variable intern_l : SIGNED (0 to L'LENGTH-1);
  variable intern_r : SIGNED (0 to R'LENGTH-1);
begin
  intern_l:=l;
  intern_r:=r;
  intern_l(0):=not intern_l(0);
  intern_r(0):=not intern_r(0);
  return bit_vector (intern_L) <= bit_vector (intern_R) ;
  end SIGNED_less_or_equal ;


--=========================Exported Functions=================================

     -- Id: A.1
function "abs" ( x : SIGNED) return SIGNED is
    constant arg_LEFT:integer := x'length-1;
    variable result : SIGNED (arg_LEFT downto 0);
begin
  if x'length<1 then return nas; end if;
  result:=x;
  if result(result'LEFT) = '1' then
    result := -result;
    end if;
  return result;
  end; -- "abs"


     -- Id: A.2
function "-" ( ARG : SIGNED) return SIGNED is
  constant arg_LEFT:integer := ARG'length-1;
  alias xarg: SIGNED(arg_LEFT downto 0) is ARG;
  variable result : SIGNED(arg_LEFT downto 0);
  variable cbit : bit := '1';
begin
  if ARG'length<1 then return nas; end if;
  for i in 0 to result'LEFT loop
    result(i) := not(xarg(i)) xor cbit;
    cbit := cbit and not(xarg(i));
    end loop;
  return result;
  end; -- "-"


--=============================================================================


     -- Id: A.3
function "+" ( L,R : UNSIGNED) return UNSIGNED is
constant L_LEFT:integer := L'length-1;
constant R_LEFT:integer := R'length-1;
constant SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return nau; end if;
  return ADD_UNSIGNED (RESIZE(L,SIZE), RESIZE(R,SIZE), '0') ;
  end;

     -- Id: A.4
function "+" ( L,R : SIGNED) return SIGNED is
constant L_LEFT:integer := L'length-1;
constant R_LEFT:integer := R'length-1;
constant SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return nas; end if;
  return ADD_SIGNED (RESIZE(L,SIZE), RESIZE(R,SIZE), '0') ;
  end;

     -- Id: A.5
function "+" ( L : UNSIGNED; R : NATURAL) return UNSIGNED is
begin
  return L + TO_UNSIGNED( R , L'length);
  end;

     -- Id: A.6
function "+" ( L : NATURAL; R : UNSIGNED) return UNSIGNED is
begin
  return TO_UNSIGNED( L , R'length) + R;
  end;

     -- Id: A.7
function "+" ( L : SIGNED; R : INTEGER ) return SIGNED is
begin
  return L + TO_SIGNED( R , L'length);
  end;

     -- Id: A.8
function "+" ( L : INTEGER; R : SIGNED) return SIGNED is
begin
  return TO_SIGNED( L , R'length) + R;
  end;

--=============================================================================

     -- Id: A.9
function "-" ( L,R : UNSIGNED) return UNSIGNED  is
constant L_LEFT:integer := L'length-1;
constant R_LEFT:integer := R'length-1;
constant SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return nau; end if;
  return ADD_UNSIGNED (RESIZE(L,SIZE), 
                     not(RESIZE(R,SIZE)), 
                       '1') ;
  end;

     -- Id: A.10
function "-" ( L,R : SIGNED) return SIGNED is
constant L_LEFT:integer := L'length-1;
constant R_LEFT:integer := R'length-1;
constant SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return nas; end if;
  return ADD_SIGNED (RESIZE(L,SIZE), 
                     not(RESIZE(R,SIZE)), 
                     '1') ;
  end;



     -- Id: A.11
function "-" ( L : UNSIGNED; R : NATURAL) return UNSIGNED is
begin
  return L -  TO_UNSIGNED( R , L'length);
  end;

     -- Id: A.12
function "-" ( L : NATURAL; R : UNSIGNED) return UNSIGNED is
begin
  return TO_UNSIGNED( L , R'length) - R;
  end;

     -- Id: A.13
function "-" ( L : SIGNED; R : INTEGER ) return SIGNED is
begin
  return L - TO_SIGNED( R , L'length);
  end;

     -- Id: A.14
function "-" ( L : INTEGER; R : SIGNED) return SIGNED is
begin
  return  TO_SIGNED( L , R'length) - R ;
  end;


--=============================================================================

     -- Id: A.15
function "*" ( L,R : UNSIGNED) return UNSIGNED is
constant L_LEFT:integer := L'length-1;
constant R_LEFT:integer := R'length-1;
alias xl : UNSIGNED(L_LEFT downto 0) is L;
alias xr : UNSIGNED(R_LEFT downto 0) is R;
variable result : UNSIGNED((L'length+R'length-1) downto 0) :=(others=>'0');
variable adval : UNSIGNED((L'length+R'length-1) downto 0);
begin
  if ((L'length<1) or (R'length<1)) then return nau; end if;
  adval := RESIZE(xr,result'length);
  for i in 0 to L_LEFT loop
    if xl(i)='1' then result := result + adval;
    end if;
    adval := shift_LEFT(adval,1);  
    end loop;
  return result;
  end;

     -- Id: A.16
function "*" ( L,R: SIGNED) return SIGNED is
  constant L_left:INTEGER:= L'length-1;
  constant R_left:INTEGER:= R'length-1;
  variable xl : SIGNED(L_left downto 0);
  variable xr : SIGNED(R_left downto 0);
  variable RESULT: SIGNED((L_Left+R_Left+1) downto 0) :=(others=>'0');
  variable adval : SIGNED((L_Left+R_Left+1) downto 0);
begin
  if ((L_Left<0) or (R_Left<0)) then return NAS; end if;
  xl := L;
  xr := R;
  adval := RESIZE(xr,RESULT'length);
  for i in 0 to L_left-1 loop
    if xl(i)='1' then RESULT:= RESULT + adval;
    end if;
    adval := shift_left(adval,1);  
    end loop;
  if xl(L_left)='1' then
    RESULT:= RESULT - adval;
    end if;
  return RESULT;
end;


     -- Id: A.17
function "*" ( L : UNSIGNED; R : NATURAL) return UNSIGNED is
begin
  return L *  TO_UNSIGNED( R , L'length);
  end;

     -- Id: A.18
function "*" ( L : NATURAL; R : UNSIGNED) return UNSIGNED is
begin
  return TO_UNSIGNED( L , R'length) * R;
  end;

     -- Id: A.19
function "*" ( L : SIGNED; R : INTEGER ) return SIGNED is
begin
  return L * TO_SIGNED( R , L'length);
  end;

     -- Id: A.20
function "*" ( L : INTEGER; R : SIGNED) return SIGNED is
begin
  return  TO_SIGNED( L , R'length) * R ;
  end;

--=============================================================================

     -- Id: A.21
function "/" ( L,R : UNSIGNED) return UNSIGNED is
variable fquot,fremain : UNSIGNED(L'length-1 downto 0);
begin
  if ((L'length<1) or (R'length<1)) then return nau; end if;
  divMod(l,r,fquot,fremain);
  return fquot;
  end;

     -- Id: A.22
function "/" ( L,R : SIGNED) return SIGNED is
variable fquot,fremain : UNSIGNED(L'length-1 downto 0);
variable xnum: UNSIGNED(L'length-1 downto 0);
variable xdenom: UNSIGNED(R'length-1 downto 0);
variable qneg: boolean := FALSE;
begin
  if ((L'length<1) or (R'length<1)) then return nas; end if;
  if L(L'LEFT)='1' then
    xnum:=UNSIGNED(-l);
    qNeg:=true;
  else
    xnum:=UNSIGNED(l);
    end if;
  if R(R'LEFT)='1' then
    xdenom:=UNSIGNED(-r);
    qNeg:=not qNeg;
  else
    xdenom:=UNSIGNED(r);
    end if;
  divMod(xnum,xdenom,fquot,fremain);
  if qNeg then fquot:="0"-fquot;  end if;
  return SIGNED(fquot);
  end;

     -- Id: A.23
function "/" ( L : UNSIGNED; R : NATURAL) return UNSIGNED is
  constant R_Length : Natural := max(L'Length, unsigned_num_bits(R));
  variable XR,Quot : unsigned (R_Length-1 downto 0);
begin
  XR := TO_UNSIGNED(R, R_Length);
  Quot := L / XR;
  if R_Length>L'Length
      and Quot(R_Length-1 downto L'Length)/=(R_Length-1 downto L'Length => '0')
  then
    ASSERT no_warning report "Numeric_bit.""/"": Quotient Truncated"
	  severity warning;
  end if;
  return Quot(L'Length-1 downto 0);
  end;

     -- Id: A.24
function "/" ( L : NATURAL; R : UNSIGNED) return UNSIGNED is
  constant L_Length : Natural := max(unsigned_num_bits(L), R'length);
  variable XL,Quot : UNSIGNED (L_Length-1 downto 0);
begin
  XL := TO_UNSIGNED(L,L_LENGTH);
  QUOT := XL / R;
  if L_LENGTH>R'LENGTH
      and QUOT(L_LENGTH-1 downto R'Length)/=(L_LENGTH-1 downto R'Length => '0')
  then
    ASSERT no_warning report "Numeric_bit.""/"": Quotient Truncated"
	  severity warning;
  end if;
  return Quot(R'Length-1 downto 0);
  end;

     -- Id: A.25
function "/" ( L : SIGNED; R : INTEGER ) return SIGNED is
  constant R_Length : Natural := max(L'Length, signed_num_bits(R));
  variable XR,Quot : signed (R_Length-1 downto 0);
begin
  XR := TO_SIGNED(R, R_Length);
  Quot := L / XR;
  if R_Length>L'Length and Quot(R_Length-1 downto L'Length)
            /= (R_Length-1 downto L'Length => Quot(L'Length-1))
  then
    ASSERT no_warning report "Numeric_bit.""/"": Quotient Truncated"
          severity warning;
  end if;
  return Quot(L'Length-1 downto 0);
  end;

     -- Id: A.26
function "/" ( L : INTEGER; R : SIGNED) return SIGNED is
  constant L_Length : Natural := max(signed_num_bits(L), R'length); 
  variable XL,Quot : SIGNED (L_Length-1 downto 0);
begin
  XL := TO_SIGNED(L,L_LENGTH);
  QUOT := XL / R;
  if L_LENGTH>R'LENGTH and QUOT(L_LENGTH-1 downto R'Length)
           /= (L_LENGTH-1 downto R'Length => Quot(R'Length-1))
  then
    ASSERT no_warning report "Numeric_bit.""/"": Quotient Truncated"
          severity warning;
  end if; 
  return Quot(R'Length-1 downto 0);
  end;

--=============================================================================

     -- Id: A.27
function "rem" ( L,R : UNSIGNED) return UNSIGNED is
variable fquot,fremain : UNSIGNED(l'length-1 downto 0);
begin
  if ((L'length<1) or (R'length<1)) then return nau; end if;
  divMod(L,R,fquot,fremain);
  return fremain;
  end;

     -- Id: A.28
function "rem" ( L,R : SIGNED) return SIGNED is
  variable fquot,fremain : UNSIGNED(l'length-1 downto 0);
  variable xnum: UNSIGNED(l'length-1 downto 0);
  variable xdenom: UNSIGNED(r'length-1 downto 0);
  variable rneg: boolean := FALSE;
begin
    if ((L'length<1) or (R'length<1)) then return nas; end if;
    if L(L'LEFT)='1' then
       xnum:=UNSIGNED(-L);
       rNeg:=true;
    else
       xNum:=UNSIGNED(L);
    end if;
    if R(R'LEFT)='1' then
       xdenom:=UNSIGNED(-R);
    else
       xdenom:=UNSIGNED(R);
    end if;
    divMod(xnum,xdenom,fquot,fremain);
    if rNeg then
       fremain:="0"-fremain;
    end if;
    return SIGNED(fremain);
  end;

     -- Id: A.29
function "rem" ( L : UNSIGNED; R : NATURAL) return UNSIGNED is
  constant R_Length : Natural := max(L'Length, unsigned_num_bits(R));
  variable XR,XRem : unsigned (R_Length-1 downto 0);
begin
  XR := TO_UNSIGNED(R, R_Length);
  XRem := L rem XR;
  if R_Length>L'Length and XRem(R_Length-1 downto L'Length)
          /= (R_Length-1 downto L'Length => '0')
  then
    ASSERT no_warning report "Numeric_bit.""rem"": Remainder Truncated"
	  severity warning;
  end if;
  return XRem(L'Length-1 downto 0);
  end;

     -- Id: A.30
function "rem" ( L : NATURAL; R : UNSIGNED) return UNSIGNED is
  constant L_Length : Natural := max(unsigned_num_bits(L), R'Length);
  variable XL,XRem : unsigned (L_Length-1 downto 0);
begin
  XL := TO_UNSIGNED(L, L_Length);
  XRem := XL rem R;
  if L_Length>R'Length and XRem(L_Length-1 downto R'Length)
          /= (L_Length-1 downto R'Length => '0')
  then
    ASSERT no_warning report "Numeric_bit.""rem"": Remainder Truncated"
	  severity warning;
  end if;
  return XRem(R'Length-1 downto 0);
  end;

     -- Id: A.31
function "rem" ( L : SIGNED; R : INTEGER ) return SIGNED is
  constant R_Length : Natural := max(L'Length, signed_num_bits(R));
  variable XR,XRem : signed (R_Length-1 downto 0);
begin
  XR := TO_SIGNED(R, R_Length);
  XRem := L rem XR;
  if R_Length>L'Length and XRem(R_Length-1 downto L'Length)
          /= (R_Length-1 downto L'Length => XRem(L'Length-1))
  then
    ASSERT no_warning report "Numeric_bit.""rem"": Remainder Truncated"
	  severity warning;
  end if;
  return XRem(L'Length-1 downto 0);
  end;

     -- Id: A.32
function "rem" ( L : INTEGER; R : SIGNED) return SIGNED is
  constant L_Length : Natural := max(signed_num_bits(L), R'Length);
  variable XL,XRem : signed (L_Length-1 downto 0);
begin
  XL := TO_SIGNED(L, L_Length);
  XRem := XL rem R;
  if L_Length>R'Length and XRem(L_Length-1 downto R'Length)
          /= (L_Length-1 downto R'Length => XRem(R'Length-1))
  then
    ASSERT no_warning report "Numeric_bit.""rem"": Remainder Truncated"
	  severity warning;
  end if;
  return XRem(R'Length-1 downto 0);
  end;

--=============================================================================

     -- Id: A.33
function "mod" ( L,R : UNSIGNED) return UNSIGNED is
  variable fquot,fremain : UNSIGNED(l'length-1 downto 0);
begin
  if ((L'length<1) or (R'length<1)) then return nau; end if;
  divMod(l,r,fquot,fremain);
  return fremain;
  end;

     -- Id: A.34
function "mod" ( L,R : SIGNED) return SIGNED is
    variable fquot,fremain : UNSIGNED(l'length-1 downto 0);
    variable xnum: UNSIGNED(l'length-1 downto 0);
    variable xdenom: UNSIGNED(r'length-1 downto 0);
    variable rneg: boolean := FALSE;
begin
    if ((L'length<1) or (R'length<1)) then return nas; end if;
    if l(l'LEFT)='1' then
       xnum:=UNSIGNED(-l);
    else
       xNum:=UNSIGNED(l);
    end if;
    if r(r'LEFT)='1' then
       xdenom:=UNSIGNED(-r);
       rNeg:=true;
    else
       xdenom:=UNSIGNED(r);
    end if;
    divMod(xnum,xdenom,fquot,fremain);
    if rNeg and l(l'LEFT)='1' then
       fremain:="0"-fremain;
    elsif rNeg then
       fremain:=fremain-xdenom;
    elsif l(l'left)='1' and fremain/="0" then
       fremain:=xdenom-fremain;
    end if;
    return SIGNED(fremain);
end;

     -- Id: A.35
function "mod" ( L : UNSIGNED; R : NATURAL) return UNSIGNED is
  constant R_Length : Natural := max(L'Length, unsigned_num_bits(R));
  variable XR,XRem : unsigned (R_Length-1 downto 0);
begin
  XR := TO_UNSIGNED(R, R_Length);
  XRem := L mod XR;
  if R_Length>L'Length and XRem(R_Length-1 downto L'Length)
          /= (R_Length-1 downto L'Length => '0')
  then
    ASSERT no_warning report "Numeric_bit.""mod"": modulus Truncated"
	  severity warning;
  end if;
  return XRem(L'Length-1 downto 0);
end;

     -- Id: A.36
function "mod" ( L : NATURAL; R : UNSIGNED) return UNSIGNED is
  constant L_Length : Natural := max(unsigned_num_bits(L), R'Length);
  variable XL,XRem : unsigned (L_Length-1 downto 0);
begin
  XL := TO_UNSIGNED(L, L_Length);
  XRem := XL mod R;
  if L_Length>R'Length and XRem(L_Length-1 downto R'Length)
          /= (L_Length-1 downto R'Length => '0')
  then
    ASSERT no_warning report "Numeric_bit.""mod"": modulus Truncated"
	  severity warning;
  end if;
  return XRem(R'Length-1 downto 0);
end;

     -- Id: A.37
function "mod" ( L : SIGNED; R : INTEGER ) return SIGNED is
  constant R_Length : Natural := max(L'Length, signed_num_bits(R));
  variable XR,XRem : signed (R_Length-1 downto 0);
begin
  XR := TO_SIGNED(R, R_Length);
  XRem := L mod XR;
  if R_Length>L'Length and XRem(R_Length-1 downto L'Length)
          /= (R_Length-1 downto L'Length => XRem(L'Length-1))
  then
    ASSERT no_warning report "Numeric_bit.""mod"": modulus Truncated"
	  severity warning;
  end if;
  return XRem(L'Length-1 downto 0);
end;

     -- Id: A.38
function "mod" ( L : INTEGER; R : SIGNED) return SIGNED is
  constant L_Length : Natural := max(signed_num_bits(L), R'Length);
  variable XL,XRem : signed (L_Length-1 downto 0);
begin
  XL := TO_SIGNED(L, L_Length);
  XRem := XL mod R;
  if L_Length>R'Length and XRem(L_Length-1 downto R'Length)
          /= (L_Length-1 downto R'Length => XRem(R'Length-1))
  then
    ASSERT no_warning report "Numeric_bit.""mod"": modulus Truncated"
	  severity warning;
  end if;
  return XRem(R'Length-1 downto 0);
end;

--=============================================================================

     -- Id: C.1
function ">"  ( L,R : UNSIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return not UNSIGNED_less_or_equal (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end ">" ;

     -- Id: C.2
function ">"  ( L,R : SIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return not SIGNED_less_or_equal (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end ">" ;

     -- Id: C.3
function ">"  ( L : NATURAL; R : UNSIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if; 
  if UnSIGNED_Num_Bits(l)>R'length then return true; end if;
  return not UNSIGNED_less_or_equal (TO_UNSIGNED (L,R'LENGTH), R) ;
end ">" ;

     -- Id: C.4
function ">"  ( L : INTEGER; R : SIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if; 
  if Signed_Num_Bits(l)>R'length then return L>0; end if;
  return not SIGNED_less_or_equal (TO_SIGNED(L,R'LENGTH), R) ;
end ">" ;

     -- Id: C.5
function ">"  ( L : UNSIGNED; R : NATURAL) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if; 
  if UnSIGNED_Num_Bits(R)>L'length then return FALSE; end if;
  return not UNSIGNED_less_or_equal (L, TO_UNSIGNED (R,L'LENGTH)) ;
end ">" ;

     -- Id: C.6
function ">"  ( L : SIGNED; R : INTEGER ) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if; 
  if Signed_Num_Bits(R)>L'length then return 0>R; end if;
  return not SIGNED_less_or_equal (L, TO_SIGNED(R,L'LENGTH)) ;
end ">" ;

--=============================================================================

     -- Id: C.7
function "<"  ( L,R : UNSIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return UNSIGNED_less (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end "<" ;

     -- Id: C.8
function "<"  ( L,R : SIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return SIGNED_less (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end "<" ;

     -- Id: C.9
function "<"  ( L : NATURAL; R : UNSIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(L)>R'length then return L<0; end if;
  return UNSIGNED_less (TO_UNSIGNED(L,R'LENGTH), R) ;
end "<" ;

     -- Id: C.10
function "<"  ( L : INTEGER; R : SIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if Signed_Num_Bits(L)>R'length then return L<0; end if;
  return SIGNED_less (TO_SIGNED(L,R'LENGTH), R) ;
end "<" ;

     -- Id: C.11
function "<"  ( L : UNSIGNED; R : NATURAL) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(R)>L'length then return 0<R; end if;
  return UNSIGNED_less (L, TO_UNSIGNED (R,L'LENGTH)) ;
end "<" ;

     -- Id: C.12
function "<"  ( L : SIGNED; R : INTEGER ) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if Signed_Num_Bits(R)>L'length then return 0<R; end if;
  return SIGNED_less (L, TO_SIGNED (R,L'LENGTH)) ;
end "<" ;

--=============================================================================

     -- Id: C.13
function "<=" ( L,R : UNSIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return UNSIGNED_less_or_equal (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end "<=" ;

     -- Id: C.14
function "<=" ( L,R : SIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return SIGNED_less_or_equal (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end "<=" ;

     -- Id: C.15
function "<=" ( L : NATURAL; R : UNSIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(L)>R'length then return L<0; end if;
  return UNSIGNED_less_or_equal (TO_UNSIGNED(L,R'LENGTH), R) ;
end "<=" ;

     -- Id: C.16
function "<=" ( L : INTEGER; R : SIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if Signed_Num_Bits(L)>R'length then return L<0; end if;
  return SIGNED_less_or_equal (TO_SIGNED(L,R'LENGTH), R) ;
end "<=" ;

     -- Id: C.17
function "<=" ( L : UNSIGNED; R : NATURAL) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(R)>L'length then return 0<R; end if;
  return UNSIGNED_less_or_equal (L, TO_UNSIGNED(R,L'LENGTH)) ;
end "<=" ;

     -- Id: C.18
function "<=" ( L : SIGNED; R : INTEGER ) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if Signed_Num_Bits(R)>L'length then return 0<R; end if;
  return SIGNED_less_or_equal (L, TO_SIGNED(R,L'LENGTH)) ;
end "<=" ;

--=============================================================================

     -- Id: C.19
function ">=" ( L,R : UNSIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return not UNSIGNED_less (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end ">=" ;

     -- Id: C.20
function ">=" ( L,R : SIGNED) return BOOLEAN is 
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return not SIGNED_less (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end ">=" ;

     -- Id: C.21
function ">=" ( L : NATURAL; R : UNSIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(L)>R'length then return L>0; end if;
  return not UNSIGNED_less (TO_UNSIGNED(L,R'LENGTH), R) ;
end ">=" ;

     -- Id: C.22
function ">=" ( L : INTEGER; R : SIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if Signed_Num_Bits(L)>R'length then return L>0; end if;
  return not SIGNED_less (TO_SIGNED (L,R'LENGTH), R) ;
end ">=" ;

     -- Id: C.23
function ">=" ( L : UNSIGNED; R : NATURAL) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(R)>L'length then return 0>R; end if;
  return not UNSIGNED_less (L, TO_UNSIGNED(R,L'LENGTH)) ;
end ">=" ;

     -- Id: C.24
function ">=" ( L : SIGNED; R : INTEGER ) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if Signed_Num_Bits(R)>L'length then return 0>R; end if;
  return not SIGNED_less (L, TO_SIGNED(R,L'LENGTH)) ; 
end ">=" ;

--=============================================================================

     -- Id: C.25
function "="  ( L,R : UNSIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return UNSIGNED_equal (RESIZE(L,SIZE),RESIZE(R,SIZE)) ;
end "=" ;

     -- Id: C.26
function "="  ( L,R : SIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return SIGNED_equal (RESIZE(L,SIZE), RESIZE(R,SIZE)) ;
end "=" ;

     -- Id: C.27
function "="  ( L : NATURAL; R : UNSIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(L)>R'Length then return FALSE; end if;
  return UNSIGNED_equal ( TO_UNSIGNED(L,R'LENGTH), R);
end "=" ;

     -- Id: C.28
function "="  ( L : INTEGER; R : SIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if Signed_Num_Bits(L)>R'Length then return FALSE; end if;
  return SIGNED_equal ( TO_SIGNED(L,R'LENGTH), R) ;
end "=" ;

     -- Id: C.29
function "="  ( L : UNSIGNED; R : NATURAL) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(R)>L'Length then return FALSE; end if;
  return UNSIGNED_equal (L, TO_UNSIGNED (R,L'LENGTH)) ;
end "=" ;

     -- Id: C.30
function "="  ( L : SIGNED; R : INTEGER ) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if Signed_Num_Bits(R)>L'Length then return FALSE; end if;
  return SIGNED_equal (L, TO_SIGNED (R,L'LENGTH)) ;
end "=" ;


--=============================================================================

     -- Id: C.31
function "/="  ( L,R : UNSIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return not(UNSIGNED_equal (RESIZE(L,SIZE),RESIZE(R,SIZE))) ;
end "/=" ;

     -- Id: C.32
function "/="  ( L,R : SIGNED) return BOOLEAN is
    VARIABLE SIZE : NATURAL := MAX (L'LENGTH, R'LENGTH) ;
begin
  if ((L'length<1) or (R'length<1)) then return FALSE; end if;
  return not(SIGNED_equal (RESIZE(L,SIZE), RESIZE(R,SIZE))) ;
end "/=" ;

     -- Id: C.33
function "/="  ( L : NATURAL; R : UNSIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(L)>R'Length then return true; end if;
  return not(UNSIGNED_equal ( TO_UNSIGNED(L,R'LENGTH), R));
end "/=" ;

     -- Id: C.34
function "/="  ( L : INTEGER; R : SIGNED) return BOOLEAN is
begin
  if (R'length<1) then return FALSE; end if;
  if Signed_Num_Bits(L)>R'Length then return true; end if;
  return not(SIGNED_equal ( TO_SIGNED(L,R'LENGTH), R)) ;
end "/=" ;

     -- Id: C.35
function "/="  ( L : UNSIGNED; R : NATURAL) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if UnSIGNED_Num_Bits(R)>L'Length then return true; end if;
  return not(UNSIGNED_equal (L, TO_UNSIGNED (R,L'LENGTH))) ;
end "/=" ;

     -- Id: C.36
function "/="  ( L : SIGNED; R : INTEGER ) return BOOLEAN is
begin
  if (L'length<1) then return FALSE; end if;
  if Signed_Num_Bits(R)>L'Length then return true; end if;
  return not(SIGNED_equal (L, TO_SIGNED (R,L'LENGTH))) ;
end "/=" ;


--=============================================================================

     -- Id: S.1
function shift_LEFT (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED is
begin
  if (ARG'length<1) then return nau; end if;
  return UNSIGNED(xsll(bit_vector(ARG),count));
end;

     -- Id: S.2
function shift_RIGHT (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED is
begin
  if (ARG'length<1) then return nau; end if;
  return UNSIGNED(xsrl(bit_vector(ARG),count));
end;

     -- Id: S.3
function shift_LEFT (  ARG  : SIGNED; count : NATURAL) return SIGNED is
begin
  if (ARG'length<1) then return nas; end if;
  return SIGNED(xsll(bit_vector(ARG),count));
end;

     -- Id: S.4
function shift_RIGHT (  ARG  : SIGNED; count : NATURAL) return SIGNED is
begin
  if (ARG'length<1) then return nas; end if;
  return SIGNED(xsra(bit_vector(ARG),count));
end;

--=============================================================================

     -- Id: S.5
function ROTATE_LEFT (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED is
begin
  if (ARG'length<1) then return nau; end if;
  return UNSIGNED(xrol(bit_vector(ARG),count));
end;

     -- Id: S.6
function ROTATE_RIGHT (  ARG  : UNSIGNED; count : NATURAL) return UNSIGNED is
begin
  if (ARG'length<1) then return nau; end if;
  return UNSIGNED(xror(bit_vector(ARG),count));
end;

     -- Id: S.7
function ROTATE_LEFT (  ARG  : SIGNED; count : NATURAL) return SIGNED is
begin
  if (ARG'length<1) then return nas; end if;
  return SIGNED(xrol(bit_vector(ARG),count));
end;

     -- Id: S.8
function ROTATE_RIGHT (  ARG  : SIGNED; count : NATURAL) return SIGNED is
begin
  if (ARG'length<1) then return nas; end if;
  return SIGNED(xror(bit_vector(ARG),count));
end;

--=============================================================================

     -- Id: D.1
function TO_INTEGER ( ARG : UNSIGNED) return NATURAL is
   constant arg_LEFT:integer := ARG'length-1;
   alias  xarg :UNSIGNED(arg_LEFT downto 0) is ARG; 
   variable result : NATURAL := 0;
   variable w : INTEGER := 1;  -- weight factor
begin
   if (ARG'length<1) then
     assert NO_WARNING
     report "numeric_std.TO_INTEGER : null arg"
     severity warning;
     return 0;
     end if;
   for i in xarg'reverse_range loop
     if xarg (i) = '1' then
       result := result + w;
       end if;
     if (i /= xarg'LEFT) then w := w + w;
       end if;
     end loop;
   return result;
end TO_INTEGER;

     -- Id: D.2
function TO_INTEGER ( ARG : SIGNED) return INTEGER is
begin

    if ARG(ARG'LEFT) = '0' then
       return TO_INTEGER( UNSIGNED (ARG)) ;
    else
       return (- (TO_INTEGER( UNSIGNED ( - (ARG + 1)))) -1);
    end if;
end TO_INTEGER;

     -- Id: D.3
function TO_UNSIGNED ( ARG,SIZE : NATURAL) return UNSIGNED is
  VARIABLE result : UNSIGNED (SIZE-1 downto 0) ;
  variable i_val:natural := ARG;
  begin
  if (size < 1) then return nau; end if;
  for i in 0 to result'LEFT loop
    if (i_val MOD 2) = 0 then
       result(i) := '0';
    else result(i) := '1' ;
      end if;
    i_val := i_val/2 ;
    end loop;
  if not(i_val=0) then
    assert NO_WARNING 
    report "NUMERIC_BIT.TO_UNSIGNED : vector truncated"
    severity WARNING ;
    end if;
  return result ;
  end TO_UNSIGNED;

     -- Id: D.4
function TO_SIGNED ( ARG : INTEGER;
                     SIZE : NATURAL) return SIGNED is
  VARIABLE result : SIGNED (SIZE-1 downto 0) ;
  VARIABLE b_val : bit := '0' ;
  VARIABLE i_val : INTEGER := ARG ;
  begin
  if (size < 1) then return nas; end if;
  if (ARG<0) then
    b_val := '1' ;
    i_val := -(ARG+1) ;
    end if ;
  for i in 0 to result'LEFT loop
    if (i_val MOD 2) = 0 then
      result(i) := b_val;
    else 
      result(i) := not b_val ;
      end if;
    i_val := i_val/2 ;
    end loop;
  if ((i_val/=0) or (b_val/=result(result'LEFT))) then
    assert NO_WARNING 
    report "NUMERIC_BIT.TO_SIGNED : vector truncated"
    severity WARNING ;
    end if;
  return result;
  end TO_SIGNED;

     -- Id: D.5
  function TO_UNSIGNED ( ARG : BIT_VECTOR) return UNSIGNED is
    begin
    return UNSIGNED(ARG);
    end TO_UNSIGNED;

     -- Id: D.6
  function TO_SIGNED ( ARG : BIT_VECTOR) return SIGNED is
    begin
    return SIGNED(ARG);
    end TO_SIGNED;

     -- Id: D.7
  function TO_BITVECTOR ( ARG : UNSIGNED) return BIT_VECTOR is
    begin
    return bit_vector(ARG);
    end TO_BITVECTOR;

     -- Id: D.8
  function TO_BITVECTOR ( ARG : SIGNED) return BIT_VECTOR is
    begin
    return bit_vector(ARG);
    end TO_BITVECTOR;

--=============================================================================


     -- Id: R.1
function RESIZE (ARG : SIGNED; new_size : NATURAL) return SIGNED is
    alias invec : SIGNED (ARG'length-1 downto 0) is ARG ;
    VARIABLE result : SIGNED (new_size-1 downto 0) ;
    constant bound : NATURAL := MIN(ARG'length,result'length)-2 ;
begin
    if (new_size<1) then return nas; end if;
    result := (others=>ARG(ARG'LEFT)) ;
    if bound >= 0 then 
      result(bound downto 0) := invec(bound downto 0) ;
      end if;
    return result;
end RESIZE ;


     -- Id: R.2
function RESIZE ( ARG : UNSIGNED; new_size : NATURAL) return UNSIGNED is
  constant ARG_LEFT:integer := ARG'length-1;
  alias XARG : UNSIGNED(ARG_LEFT downto 0) is ARG;
  variable result : UNSIGNED(new_size-1 downto 0) := (others=>'0');
begin
  if (new_size<1) then return nau; end if;
  if XARG'length=0 then return result; end if;
  if (result'length < ARG'length) then
    result(result'LEFT downto 0) := xarg(result'LEFT downto 0);
  else
    result(result'LEFT downto XARG'LEFT+1) := (others => '0');
    result(XARG'LEFT downto 0) := XARG;
    end if;
  return result;
  end RESIZE;

--=============================================================================

     -- Id: E.1
    function RISING_EDGE  (signal s : bit) return BOOLEAN is
    begin
        return s'EVENT and s = '1';
    end;

     -- Id: E.2
    function FALLING_EDGE (signal s : bit) return BOOLEAN is
    begin
        return s'EVENT and s = '0';
    end;
				       
--=============================================================================
end NUMERIC_BIT;




