#!perl
#
# Author: Bassam Tabbara, UC Berkeley (tbassam@eecs.berkeley.edu)
# SHIFT to VHDL Translator
#
# REQUIRES: running "s2s <SHIFT file>" first
#
# Usage: s2v <SHIFT file> 
# Output: file.vhd (VHDL file)
#
# Goal: output a synthesizable RTL VHDL description of SHIFT design.
#
# Strategy: generate an entity for the main net, and communicating
# processes for CFSMs. The description will be "flat" i.e. no
# hierarchy of components (CFSMs) is generated, just the entity, 
# and architecture, and communicating processes (CFSMs).
#
# Known limitations:
# 1) does not handle multiple (shared) instances i.e. they need to be
#    replicated
# 2) does not initialize BIT_VECTOR (i.e. SIGNED/UNSIGNED) bits (in SHIFT
#    the value is given in integer, and the script does not bother (yet !)
#    to do the proper encoding...
# 3) can handle one level of hierarchy correctly (i.e. one main net, 
#    with multiple subckts of CFSMs, more than one level of 
#    hierarchy is dubious...
#
# Code Rating: RED (many iteration loops should really be subroutines...)
#
# Quality: it does the job !

#get filename
$filename = sprintf "%s", $ARGV[0];
@names = split(/\./,$filename);
$name = $names[0];

#global variable declarations
$entity = 1;
$cfsm_net = 1;
$first = 1;
$transition = 0;

#subroutines

#output files
$outputfilename = sprintf ">./%s.vhd", $name;
open(OUT, $outputfilename) || die 
    "Can't open file $outputfilename: $!\n";

print OUT "--RTL VHDL description\n";
print OUT "--Automatically generated by SHIFT2VHDL\n";
print OUT "--Comments/Questions can be directed to: \n";
print OUT "\t\t--Bassam Tabbara\n\t\t--tbassam\@eecs.berkeley.edu\n\n";
print OUT "--An implementation of a network of CFSMs as communicating\n";
print OUT "--processes which act as Moore-type FSMs\n\n";
print OUT "--The description is 'flat' i.e. no\n";
print OUT "--hierarchy of components (CFSMs) is generated, just the entity\n"; 
print OUT "--and architecture, and communicating processes (CFSMs).\n\n";
print OUT "--REQUIRE: numeric_bit.vhd, routines.vhd\n\n"; 
print OUT "--include packages in design\n";
print OUT "use WORK.numeric_bit.all;\n";
print OUT "use WORK.uC_routines_synth_pkg.all;\n\n";

#parse SHIFT file
parse:
while(<>) {

    if ($entity) { #within entity/architecture
	if (/^\.net(\s+)(\w+)/) {
	    printf OUT "Entity %s is\n", $2;
	    print OUT "  port(\n";
	    print OUT "\tclk : in BIT;\n"; #clk
	}
	
	if (/^\.inputs(\s+)(.*)/) {
	    $buf = $2;
	    $buf =~ s/\s+//g;
	    @net_inputs = split(/,/, $buf);
	    
	    @net_einputs = grep(/\*/,@net_inputs); #event
	    @net_vinputs = grep(/\(/,@net_inputs); #value
	    
	    #print events (entity input)
	    foreach $input (@net_einputs) {
		$input =~ s/\*//;
		$input =~ s/\(.*\)//;
		printf OUT "\te_%s : in BIT;\n", $input;
	    }	    
	}

	if (/^\.outputs(\s+)(.*)/) {
	    $buf = $2;
	    $buf =~ s/\s+//g;
	    @net_outputs = split(/,/, $buf);
	    
	    @net_eoutputs = grep(/\*/,@net_outputs); #event
	    @net_voutputs = grep(/\(/,@net_outputs); #value
	    
	    #print events (entity output)
	    foreach $input (@net_eoutputs) {
		$input =~ s/\*//;
		$input =~ s/\(.*\)//;
	    
		printf OUT "\te_%s : buffer BIT;\n", $input;
	    }	    
	}
	
	if (/^\.nb\s+(\w+)\s+(\d+)/) { #SIGNED
	    $sbitnum{$1} = $2; #bitnum downto 0
	}

	if (/^\.nb\s+(\w+)\s+unsigned\s+(\d+)/) { #UNSIGNED
	    $bitnum{$1} = $2-1; # bitnum-1 downto 0
	}

	if (/^\.mv\s+(\w+)\s+(\d+)/) { #SIGNED
	    $sbitnum{$1} = log($2)/log(2); #bitnum downto 0
	}

	if (/^\.mv\s+(\w+)\s+unsigned\s+(\d+)/) { #UNSIGNED
	    $bitnum{$1} = log($2-1)/log(2); # bitnum-1 downto 0
	}

	if (/^\.const\s+(\w+)\s+(\d+)/) {
	    $consts{$1} = $2;
	}

	if (/^\.user_const\s+(\w+)\s+(\d+)/) {
	    $consts{$1} = $2;
	}
	
	if (/^\.internal(\s+)(.*)/) {
	 
	    $buf = $2;
	    $buf =~ s/\s+//g;
	    @net_internal = (@net_internal, split(/,/, $buf));
	}
	
	if (/^\.subckt net_(\w+) \w+\d+ (.*)/) {
	    # formal/actual (net/cfsm) association
	    $the_cfsm = $1;
	    #identify input and output matching
	    @io = split(/ \| /, $2);
	    #print "DEBUG: ", $io[0], "\n";
	    #print "DEBUG: ", $io[1], "\n";
	    @imatch = split(/[\=| ]/, $io[0]); 
	    @omatch = split(/[\=| ]/, $io[1]); 
	
	    #input matching
	    $i = 0;
	    while ($i < $#imatch) {
		$imatch[$i] =~ s/\*//;
		$imatch[$i+1] =~ s/\*//;
		$entry = $the_cfsm . $imatch[$i]; #use name of cfsm as well
		$parent{$entry} = $imatch[$i + 1];
		$i = $i + 2;
	    }
	    #output matching
	    $i = 0;
	    while ($i < $#omatch) {
		$omatch[$i] =~ s/\*//;
		$omatch[$i+1] =~ s/\*//;
		$entry = $the_cfsm . $omatch[$i]; #use name of cfsm as well
		$parent{$entry} = $omatch[$i + 1];
		$i = $i + 2;	    
	    }
	    
	}

	if (/^\.end/) {
	    $entity = 0;	    
	    #print values (entity input)
	    foreach $input (@net_vinputs) {
		$the_input = $input;
		$input =~ s/\*//;
		if ($the_input ne $input) { #value with event
		    $input =~ s/\(\w+\)//;
		}
		else { #value only
		    $input =~ s/\(//;
		    $input =~ s/\)//;
		}

		if ($first) {
		    $first = 0;
		}
		else {
		    printf OUT ";\n";
		}
		if ($bitnum{$input} ne "") {
		    printf OUT "\tv_%s : in UNSIGNED(%s downto 0)", 
		    $input, $bitnum{$input};
		}
		else {
		    if ($sbitnum{$input} ne "") {
			printf OUT "\tv_%s : in SIGNED(%s downto 0)", $input, $sbitnum{$input};
		    }
		    else  { #default
			printf OUT "\tv_%s : in integer := 0", $input;
			
		    }
		}	
	    }
	    #print values (entity output)
	    foreach $output (@net_voutputs) {
		$the_output = $output;
		$output =~ s/\*//;
		if ($the_output ne $output) { #value with event
		    $output =~ s/\(\w+\)//;
		}
		else { #value only
		    $output =~ s/\(//;
		    $output =~ s/\)//;
		}

		if ($first) {
		    $first = 0;
		}
		else {
		    printf OUT ";\n";
		}
		if ($bitnum{$output} ne "") {
		    printf OUT "\tv_%s : buffer UNSIGNED(%s downto 0)", $output, $bitnum{$output};
		}
		else {
		    if ($sbitnum{$output} ne "") {
			printf OUT "\tv_%s : buffer SIGNED(%s downto 0)", $output, $sbitnum{$output};
		    }
		    else { #default
			printf OUT "\tv_%s : buffer integer := 0", $output;
		    }
		}
	    }
	    printf OUT "\n\t);\n";
	    printf OUT "end %s;\n", $name;
	    print OUT "--include packages before architecture\n";
	    print OUT "use WORK.numeric_bit.all;\n";
	    print OUT "use WORK.uC_routines_synth_pkg.all;\n\n";
	    printf OUT "\nArchitecture CFSM_rtl of %s ", $name;
	    printf OUT "is\n";

	    @net_eint = grep(/\*/,@net_internal);
	    @net_vint = grep(/\(/,@net_internal);

	    #internals (event)
	    foreach $eint (@net_eint) {
		$eint =~ s/\*//;
		$eint =~ s/\(\w+\)//;
		printf OUT "\tsignal e_%s : BIT := '0';\n", $eint;
	    }
	    #internals (values)
	    foreach $vint (@net_vint) {
		$vint =~ s/\*\w+//;
		$vint =~ s/\((\w+)\)/$1/;		
		if ($bitnum{$vint} ne "") {
		    if ($consts{$vint} ne "") {
			printf OUT "\tsignal v_%s : UNSIGNED(%s downto 0); -- := %s\n",
			$vint, $bitnum{$vint}, $consts{$vint};
		    }
		    else { 
			printf OUT "\tsignal v_%s : UNSIGNED(%s downto 0);\n",
			$vint, $bitnum{$vint};
		    }
		}
		else  {
		    if ($sbitnum{$vint} ne "") {
			if ($consts{$vint} ne "") {
			    printf OUT "\tsignal v_%s : SIGNED(%s downto 0); --:= %s\n",
			    $vint, $sbitnum{$vint}, $consts{$vint};
			}
			else {
			    printf OUT "\tsignal v_%s : SIGNED(%s downto 0);\n",
			    $vint, $sbitnum{$vint};
			}
		    }
		    else  { #default
			if ($consts{$vint} eq '') {
			    printf OUT "\tsignal v_%s : integer := 0;\n", $vint;
			}
			else {
			    printf OUT "\tsignal v_%s : integer := %d;\n", $vint, $consts{$vint};
			}
		    }
		}
	    }
	    printf OUT "Begin\n";
	}
    }

    if (!$entity) {
	if ($cfsm_net) {	
	    #cfsm net
	    if (/^\.net net_(\w+)/) {
		$name_of_cfsm = $1;
		#reset arrays and lists
		@cfsm_internal = "";
		$cbitnum ="";
		$csbitnum = "";
		$cconsts = "";
		$oldstate = "";
		$is_ievent = "";
		$is_event = "";
		@transition_inputs = "";
		@trans_inputs = "";
		@transition_outputs = "";
		@trans_outputs = "";

		printf OUT "\n%s:\nprocess (clk)\n", $name_of_cfsm;
	    }
	    
	    #cfsm declarations
	    if (/^\.inputs(\s+)(.*)/) {
		$buf = $2;
		$buf =~ s/\s+//g;
		@cfsm_inputs = split(/,/, $buf);
		
		@cfsm_einputs = grep(/\*/,@cfsm_inputs); #event
		@cfsm_vinputs = grep(/\(/,@cfsm_inputs); #value
	    }
	    
	    if (/^\.outputs(\s+)(.*)/) {
		$buf = $2;
		$buf =~ s/\s+//g;
		@cfsm_outputs = split(/,/, $buf);
		
		@cfsm_eoutputs = grep(/\*/,@cfsm_outputs); #event
		@cfsm_voutputs = grep(/\(/,@cfsm_outputs); #value
	    }
	    
	    if (/^\.internal(\s+)(.*)/) {
		
		$buf = $2;
		$buf =~ s/\s+//g;
		@cfsm_internal = (@cfsm_internal, split(/,/, $buf));
	    }
	    	    
	    if (/^\.nb\s+(\w+)\s+(\d+)/) { #SIGNED
		$csbitnum{$1} = $2; #bitnum downto 0
	    }
	    
	    if (/^\.nb\s+(\w+)\s+unsigned\s+(\d+)/) { #UNSIGNED
		$cbitnum{$1} = $2-1; #bitnum-1 downto 0
	    }
	    
	    if (/^\.mv\s+(\w+)\s+(\d+)/) { #SIGNED
		$sbitnum{$1} = log($2)/log(2); #bitnum downto 0
	    }
	    
	    if (/^\.mv\s+(\w+)\s+unsigned\s+(\d+)/) { #UNSIGNED
		$bitnum{$1} = log($2-1)/log(2); # bitnum-1 downto 0
	    }

	    if (/^\.const\s+(\w+)\s+(\d+)/) {
		$cconsts{$1} = $2;
	    }
	    
	    if (/^\.user_const\s+(\w+)\s+(\d+)/) {
		$cconsts{$1} = $2;
	    }

	    if (/^.subckt (\w+) (\w+\d+) (.*)/) {
		if ($1 ne $name_of_cfsm) {
		    @subckt_io = split(/\|/, $3);
		    @isub = split(/[\=| ]/, $subckt_io[0]); 
		    @osub = split(/[\=| ]/, $subckt_io[1]); 
		    $i = 0;
		    while ($i < $#isub) {
			if ($i == 0) { 
			    $buf = sprintf "\tv_%s_tmp := %s(v_%s_tmp",$osub[2],$1,$isub[$i+1];
			    $save = $buf;
			}
			else {
			    $buf = sprintf ", v_%s_tmp", $isub[$i+1];
			    $save = $save . $buf;
			}
			$i = $i + 2;
		    }
		    $save = $save . ");\n"; 
		    
		    @subckts = ($save, @subckts); #reverse the order
		}
	    }

	    if (/^\.end/) {
		
	    } # of .end
	}
    }
    #cfsm body
    if (/^\.cfsm (\w+)/) {
	$cfsm_net = 0;
	$cfsm_name = $1;
	$in_cfsm = 1;

    }
    
    if ($in_cfsm) {
	
	if (/^\.inputs (.*)/) {
	    $buf = $1;
	    $buf =~ s/\s+//g;
	    $buf =~ s/\)//g;
	    @transition_inputs = split(/[,|\(]/, $buf);

	    $count = 0;
	    for ($i = 0; $i <= $#transition_inputs; $i++) {
		if ($transition_inputs[$i] ne "") {
		    $trans_inputs[$count] = $transition_inputs[$i];
		    $count++;
		}
	    }

	    foreach $input (@trans_inputs) {
		$the_input = $input;
		$the_input =~ s/\*//;
		if ($the_input ne $input) {
		    $is_ievent{$input} = 1;
		}
		else {
		    $is_ievent{$input} = 0;
		}

	    }
	}

	if (/^\.outputs (.*)/) {
	    $buf = $1;
	    $buf =~ s/\s+//g;
	    $buf =~ s/o_(\w+)/$1/g;
	    $buf =~ s/\)//g;
	    
	    @transition_outputs = split(/[,|\(]/, $buf);
	    
	    $count = 0;
	    for ($i = 0; $i <= $#transition_outputs; $i++) {
		if ($transition_outputs[$i] ne "") {
		    $trans_outputs[$count] = $transition_outputs[$i];
		    $count++;
		}
	    }
	    	    
	    foreach $output (@trans_outputs) {
		$the_output = $output;
		$the_output =~ s/\*//;
		if ($output ne $the_output) { #event
		    $is_event{$output} = 1;
		}
		else {
		    $is_event{$output} = 0;
		}
	    }
	}

	if (/^\.mv _st (\d+) (.*)/) {
	    print OUT "\ttype S_Type is (";
	    @states = split(/ /, $2);
	    $first = 1;
	    foreach $state (@states) {
		if (!$first) { print OUT ","; }
		else { $first = 0; };
		printf OUT "S%s", $state;
	    }
	    print OUT ");\n";
	}

	if (/^\.r _st (\d+)/) {
	    $reset_state = $1;
	    printf OUT "\tvariable State, Next_State : S_Type := S%s;\n", $1;
	}
	
	if (/^.r \*o_(\w+) (\d+)/) {
	    $reset{$1} = $2;
	}

	if (/^\.trans/) {
	
	    #declare events (cfsm input)
	    foreach $input (@cfsm_einputs) {
		$input =~ s/\*//;
		$input =~ s/\(\w+\)//;
		printf OUT "\tvariable e_%s_tmp : BIT;\n", $input;
	    }	

	    #print events (entity output)
	    foreach $output (@cfsm_eoutputs) {
		$output =~ s/\*//;
		$output =~ s/\(\w+\)//;
		
		#add RESETs
		if ($reset{$output} ne "") {
		    printf OUT "\tvariable e_%s_tmp : BIT := '%d';\n", $output,
		    $reset{$1};
		}
		else {
		    printf OUT "\tvariable e_%s_tmp : BIT;\n", $output;
		}	
	    }

	    #print values (cfsm input)
		foreach $input (@cfsm_vinputs) {
		    $the_input = $input;
		    $input =~ s/\*//;
		    if ($the_input ne $input) { #value with event
			$input =~ s/\(\w+\)//;
		    }
		    else { #value only
			$input =~ s/\(//;
			$input =~ s/\)//;
		    }
		    if ($cbitnum{$input} ne "") {
			printf OUT "\tvariable v_%s_tmp : UNSIGNED(%s downto 0);\n", 
			$input, $cbitnum{$input};
		    }
		    else {
			if ($csbitnum{$input} ne "") {
			    printf OUT "\tvariable v_%s_tmp : SIGNED(%s downto 0);\n", $input, $csbitnum{$input};
			}
			else  { #default
			    if ($cconsts{$input} eq '') {
				printf OUT "\tvariable v_%s_tmp : integer := 0;\n", $input;
			    }
			    else {
				printf OUT "\tvariable v_%s_tmp : integer := %d;\n", $input, $cconsts{$input};
			    }
			}
		    }	
		}
		#print values (cfsm output)
		foreach $output (@cfsm_voutputs) {
		    $the_output = $output;
		    $output =~ s/\*//;
		    if ($the_output ne $output) { #value with event
			$output =~ s/\(\w+\)//;
		    }
		    else { #value only
			$output =~ s/\(//;
			$output =~ s/\)//;
		    }

		    if ($cbitnum{$output} ne "") {
			printf OUT "\tvariable v_%s_tmp : UNSIGNED(%s downto 0);\n", $output, $cbitnum{$output};
		    }
		    else {
			if ($csbitnum{$output} ne "") {
			    printf OUT "\tvariable v_%s_tmp : SIGNED(%s downto 0);\n", $output, $csbitnum{$output};
			}
			else { #default
			    printf OUT "\tvariable v_%s_tmp : integer := 0;\n", $output;
			}
		    }
		}
		
		@cfsm_eint = grep(/\*/,@cfsm_internal);
		@cfsm_vint = grep(/\(/,@cfsm_internal);
		
		#internals (event)
		foreach $eint (@cfsm_eint) {
		    $eint =~ s/\*//;
		    printf OUT "\tvariable e_%s_tmp : bit := '0';\n", $eint;
		}
		#internals (values)
		foreach $vint (@cfsm_vint) {
		    $vint =~ s/\((\w+)\)/$1/;	
		    if ($cbitnum{$vint} ne "") {
			if ($cconsts{$vint} ne "") {
			    printf OUT "\tvariable v_%s_tmp : UNSIGNED(%s downto 0); -- := %s\n",
			    $vint, $cbitnum{$vint}, $cconsts{$vint};
			}
			else { 
			    printf OUT "\tvariable v_%s_tmp : UNSIGNED(%s downto 0);\n",
			    $vint, $cbitnum{$vint};
			}
		    }
		    else  {
			if ($csbitnum{$vint} ne "") {
			    if ($cconsts{$vint} ne "") {
				printf OUT "\tvariable v_%s_tmp : SIGNED(%s downto 0); --:= %s\n",
				$vint, $csbitnum{$vint}, $cconsts{$vint};
			    }
			    else {
				printf OUT "\tvariable v_%s_tmp : SIGNED(%s downto 0);\n",
				$vint, $csbitnum{$vint};
			    }
			}
			else  { #default
			    printf OUT "\tvariable v_%s_tmp : integer := 0;\n", $vint;
			}
		    }
		}
	    
	    print OUT "\nBegin\n";
	    print OUT "\nif ( clk'event and (clk = '1') ) then\n";
	    print OUT "\t--sample inputs\n";
	    if ($#cfsm_einputs == -1) {
		printf "WARNING: no input events found for CFSM '%s' !\n",
		$cfsm_name;
		printf "\t Are you sure you have a NETwork of CFSMs\n";
		printf "\t with ONE level of hierarchy ?\n\t..continuing...\n";
	    }
	    foreach $input (@cfsm_einputs) {
		$input =~ s/\*//;
		$entry = $cfsm_name . $input;
		printf OUT "\te_%s_tmp := e_%s;\n", $input,
		$parent{$entry};
	    }	
	    if ($#cfsm_vinputs == -1) {
		printf "WARNING: no input values found for CFSM '%s' !\n",
		$cfsm_name;
	    }
	    foreach $input (@cfsm_vinputs) {
		$the_input = $input;
		$input =~ s/\*//;
		if ($the_input ne $input) { #value with event
		    $input =~ s/\(\w+\)//;
		}
		else { #value only
		    $input =~ s/\(//;
		    $input =~ s/\)//;
		}
		$entry = $cfsm_name . $input;
		printf OUT "\tv_%s_tmp := v_%s;\n", $input,
		$parent{$entry};
	    }

	    print OUT "\t--emit outputs\n";
	    if ($#cfsm_voutputs == -1) {
		printf "WARNING: no valued outputs found for CFSM '%s' !\n",
		$cfsm_name;
	    }
	    foreach $output (@cfsm_voutputs) {
		$the_output = $output;
		$output =~ s/\*//;
		if ($the_output ne $output) { #value with event
		    $output =~ s/\(.*\)//;
		}
		else { #value only
		    $output =~ s/\(//;
		    $output =~ s/\)//;
		}
		$entry = $cfsm_name . $output;
		printf OUT "\tv_%s <= v_%s_tmp;\n",
		$parent{$entry}, $output;
	    }
	    if ($#cfsm_eoutputs == -1) {
		printf "WARNING: no event outputs found for CFSM '%s' !\n",
		$cfsm_name;
	    }
	    foreach $output (@cfsm_eoutputs) {
		$output =~ s/\*//;
		$entry = $cfsm_name . $output;
		printf OUT "\te_%s <= e_%s_tmp;\n",
		$parent{$entry}, $output;
	    }	
	    
	    print OUT "\t--update subckts\n";
	    foreach $subck (@subckts) {
		print OUT $subck;
	    }
	    
	    print OUT "\t--update state\n";
	    print OUT "\tState := Next_State;\n";
	    
	    	
	    #start TREL
	    printf OUT "\ncase State is\n";
	    $transition = 1;
	}
	
	if ($transition && !(/^\.end/) && !(/^\.trans/)) {
	    chop; #remove \n
	    $the_line = $_;
	    @line = split(/ /, $the_line);

	    $state = @line[$#trans_inputs+1];
	    if ($state ne $old_state) { #new state
		printf OUT "\twhen S%s \=\>\n", @line[$#trans_inputs+1];
		$old_state = $state;
	    }
	    
	    $found = 0;
	    for ($i = 0; $i <= $#trans_inputs; $i++) {
		if ($line[$i] ne "-") {
		    $found = 1;
		    goto foundsomething;
		}
	    }

foundsomething:
	    # input conditions
	    if ($found) {
		for ($i = 0; $i <= $#trans_inputs; $i++) {
		    $temp = $trans_inputs[$i] . "_tmp";		    
		    if ($is_ievent{$trans_inputs[$i]}) {
			$temp =~ s/\*//;
			$temp = "e_" . $temp; #bit
		    }
		    else {
			$temp = "v_" . $temp; #unsigned
		    }
		    
		    if ($i == 0) { 
			if ($line[$i] ne "-") {
			    if ($is_ievent{$trans_inputs[$i]}) {
				printf OUT "\t\tif ( (%s = '%s') ", $temp, $line[$i];
			    }
			    else {
				printf OUT "\t\tif ( (%s = \"%s\") ", $temp, $line[$i];
			    }
			    $flag = 1;
			}
			else {
			    printf OUT "\t\tif ( ";
			    $flag = 0;
			}
		    }
		    else {
			if ($line[$i] ne "-") {
			    if ($flag) {
				if ($is_ievent{$trans_inputs[$i]}) {
				    printf OUT "and (%s = '%s') ", $temp, $line[$i];
				}
				else {
				    printf OUT "and (%s = \"%s\") ", $temp, $line[$i];
				}
			    }
			    else {
				if ($is_ievent{$trans_inputs[$i]}) {
				    printf OUT " (%s = '%s') ", $temp, $line[$i];
				}
				else {
				    printf OUT " (%s = \"%s\") ", $temp, $line[$i];
				}
				$flag = 1;
			    }
			}
		    }
		}
		print OUT ") then \n";
	    }
	    
	    #outputs
	    for ($i = 0; $i <= $#trans_outputs; $i++) {
		if ($line[$i+$#trans_inputs+1 +1] ne "-") {
		    if ($is_event{$trans_outputs[$i]}) {
			$temp = $trans_outputs[$i];
			$temp =~ s/\*//;
			printf OUT "\t\t\te_%s_tmp := '%s';\n", 
			$temp, $line[$i+$#trans_inputs+1 +1];
		    }
		    else {
			$right = $line[$i+$#trans_inputs+1 +1];
			$the_right = $right;
			$the_right =~ s/\(//;
			$the_right =~ s/\)//;
			if ($the_right ne $right) {
			    printf OUT "\t\t\tv_%s_tmp := v_%s_tmp;\n", 
			    $trans_outputs[$i], $the_right;
			}
			else {
			    printf OUT "\t\t\tv_%s_tmp := %s;\n", 
			    $trans_outputs[$i], $right;
			}
		    }
		}
	    }
	    
	    printf OUT "\t\t\tNext_State := S%s;\n", $line[$#trans_inputs+1 +1 +$#trans_outputs+1];
	    #closing
	    if ($found) {
		print OUT "\t\tend if;\n";
	    }
	}

	if (/^\.end/) {
	    print OUT "end case;\n\n";
	    printf OUT "end if;\n\n";
	    printf OUT "end process %s;\n\n", $cfsm_name;
	    $in_cfsm = 0;
	    $cfsm_net = 1;
	    $transition = 0;
	}

    }
    
}

print OUT "end CFSM_rtl;\n";

#done
close(OUT);


